from __future__ import annotations

import atexit
import contextlib
import datetime
import re
from typing import TYPE_CHECKING, Any
from urllib.parse import parse_qs, urlparse

import pyexasol
import sqlglot as sg
import sqlglot.expressions as sge

import ibis
import ibis.common.exceptions as com
import ibis.expr.datatypes as dt
import ibis.expr.operations as ops
import ibis.expr.schema as sch
import ibis.expr.types as ir
from ibis import util
from ibis.backends.base.sqlglot import SQLGlotBackend
from ibis.backends.base.sqlglot.compiler import STAR, C
from ibis.backends.exasol.compiler import ExasolCompiler

if TYPE_CHECKING:
    from collections.abc import Iterable, Mapping

    import pandas as pd
    import pyarrow as pa

    from ibis.backends.base import BaseBackend

# strip trailing encodings e.g., UTF8
_VARCHAR_REGEX = re.compile(r"^((VAR)?CHAR(?:\(\d+\)))?(?:\s+.+)?$")


class Backend(SQLGlotBackend):
    name = "exasol"
    compiler = ExasolCompiler()
    supports_temporary_tables = False
    supports_create_or_replace = False
    supports_in_memory_tables = False
    supports_python_udfs = False

    @property
    def version(self) -> str:
        # https://stackoverflow.com/a/67500385
        query = (
            sg.select("param_value")
            .from_(sg.table("EXA_METADATA", catalog="SYS"))
            .where(C.param_name.eq("databaseProductVersion"))
        )
        with self._safe_raw_sql(query) as result:
            [(version,)] = result.fetchall()
        return version

    def do_connect(
        self,
        user: str,
        password: str,
        host: str = "localhost",
        port: int = 8563,
        **kwargs: Any,
    ) -> None:
        """Create an Ibis client connected to an Exasol database.

        Parameters
        ----------
        user
            Username used for authentication.
        password
            Password used for authentication.
        host
            Hostname to connect to (default: "localhost").
        port
            Port number to connect to (default: 8563)
        kwargs
            Additional keyword arguments passed to `pyexasol.connect`.

        """
        if kwargs.pop("quote_ident", None) is not None:
            raise com.UnsupportedArgumentError(
                "Setting `quote_ident` to anything other than `True` is not supported. "
                "Ibis requires all identifiers to be quoted to work correctly."
            )

        self.con = pyexasol.connect(
            dsn=f"{host}:{port}",
            user=user,
            password=password,
            quote_ident=True,
            **kwargs,
        )

    def _from_url(self, url: str, **kwargs) -> BaseBackend:
        """Construct an ibis backend from a URL."""
        url = urlparse(url)
        query_params = parse_qs(url.query)
        kwargs = {
            "user": url.username,
            "password": url.password,
            "schema": url.path[1:] or None,
            "host": url.hostname,
            "port": url.port,
        } | kwargs

        for name, value in query_params.items():
            if len(value) > 1:
                kwargs[name] = value
            elif len(value) == 1:
                kwargs[name] = value[0]
            else:
                raise com.IbisError(f"Invalid URL parameter: {name}")

        self._convert_kwargs(kwargs)

        return self.connect(**kwargs)

    @contextlib.contextmanager
    def begin(self):
        # pyexasol doesn't have a cursor method
        con = self.con
        try:
            yield con
        except Exception:
            con.rollback()
            raise
        else:
            con.commit()

    @contextlib.contextmanager
    def _safe_raw_sql(self, query: str, *args, **kwargs):
        with contextlib.suppress(AttributeError):
            query = query.sql(dialect=self.dialect)

        with self.begin() as cur:
            yield cur.execute(query, *args, **kwargs)

    def list_tables(self, like=None, database=None):
        tables = sg.select("table_name").from_(
            sg.table("EXA_ALL_TABLES", catalog="SYS")
        )
        views = sg.select(sg.column("view_name").as_("table_name")).from_(
            sg.table("EXA_ALL_VIEWS", catalog="SYS")
        )

        if database is not None:
            tables = tables.where(sg.column("table_schema").eq(sge.convert(database)))
            views = views.where(sg.column("view_schema").eq(sge.convert(database)))

        query = sg.union(tables, views)

        with self._safe_raw_sql(query) as con:
            tables = con.fetchall()

        return self._filter_with_like([table for (table,) in tables], like=like)

    def get_schema(
        self, table_name: str, schema: str | None = None, database: str | None = None
    ) -> sch.Schema:
        name_type_pairs = self._metadata(
            sg.select(STAR)
            .from_(
                sg.table(
                    table_name, db=schema, catalog=database, quoted=self.compiler.quoted
                )
            )
            .sql(self.dialect)
        )
        return sch.Schema.from_tuples(name_type_pairs)

    def _fetch_from_cursor(self, cursor, schema: sch.Schema) -> pd.DataFrame:
        import pandas as pd

        from ibis.backends.exasol.converter import ExasolPandasData

        df = pd.DataFrame.from_records(cursor, columns=schema.names, coerce_float=True)
        df = ExasolPandasData.convert_table(df, schema)
        return df

    def _metadata(self, query: str) -> Iterable[tuple[str, dt.DataType]]:
        table = sg.table(util.gen_name("exasol_metadata"), quoted=self.compiler.quoted)
        dialect = self.dialect
        create_view = sg.exp.Create(
            kind="VIEW",
            this=table,
            expression=sg.parse_one(query, dialect=dialect),
        )
        drop_view = sg.exp.Drop(kind="VIEW", this=table)
        describe = sg.exp.Describe(this=table)
        with self._safe_raw_sql(create_view):
            try:
                yield from (
                    (
                        name,
                        self.compiler.type_mapper.from_string(
                            _VARCHAR_REGEX.sub(r"\1", typ)
                        ),
                    )
                    for name, typ, *_ in self.con.execute(
                        describe.sql(dialect=dialect)
                    ).fetchall()
                )
            finally:
                self.con.execute(drop_view.sql(dialect=dialect))

    def _register_in_memory_table(self, op: ops.InMemoryTable) -> None:
        schema = op.schema
        if null_columns := [col for col, dtype in schema.items() if dtype.is_null()]:
            raise com.IbisTypeError(
                "Exasol cannot yet reliably handle `null` typed columns; "
                f"got null typed columns: {null_columns}"
            )

        # only register if we haven't already done so
        if (name := op.name) not in self.list_tables():
            quoted = self.compiler.quoted
            column_defs = [
                sg.exp.ColumnDef(
                    this=sg.to_identifier(colname, quoted=quoted),
                    kind=self.compiler.type_mapper.from_ibis(typ),
                    constraints=(
                        None
                        if typ.nullable
                        else [
                            sg.exp.ColumnConstraint(
                                kind=sg.exp.NotNullColumnConstraint()
                            )
                        ]
                    ),
                )
                for colname, typ in schema.items()
            ]

            ident = sg.to_identifier(name, quoted=quoted)
            create_stmt = sg.exp.Create(
                kind="TABLE",
                this=sg.exp.Schema(this=ident, expressions=column_defs),
            )
            create_stmt_sql = create_stmt.sql(self.name)

            df = op.data.to_frame()
            data = df.itertuples(index=False, name=None)

            def process_item(item: Any):
                """Handle inserting timestamps with timezones."""
                if isinstance(item, datetime.datetime):
                    if item.tzinfo is not None:
                        item = item.tz_convert("UTC").tz_localize(None)
                    return item.isoformat(sep=" ", timespec="milliseconds")
                return item

            rows = (tuple(map(process_item, row)) for row in data)
            with self._safe_raw_sql(create_stmt_sql):
                if not df.empty:
                    self.con.ext.insert_multi(name, rows)

            atexit.register(self._clean_up_tmp_table, ident)

    def _clean_up_tmp_table(self, ident: sge.Identifier) -> None:
        with self._safe_raw_sql(
            sge.Drop(kind="TABLE", this=ident, force=True, cascade=True)
        ):
            pass

    def create_table(
        self,
        name: str,
        obj: pd.DataFrame | pa.Table | ir.Table | None = None,
        *,
        schema: sch.Schema | None = None,
        database: str | None = None,
        overwrite: bool = False,
        temp: bool = False,
    ) -> ir.Table:
        """Create a table in Snowflake.

        Parameters
        ----------
        name
            Name of the table to create
        obj
            The data with which to populate the table; optional, but at least
            one of `obj` or `schema` must be specified
        schema
            The schema of the table to create; optional, but at least one of
            `obj` or `schema` must be specified
        database
            The database in which to create the table; optional
        overwrite
            If `True`, replace the table if it already exists, otherwise fail
            if the table exists
        temp
            Create a temporary table (not supported)

        """
        if obj is None and schema is None:
            raise ValueError("Either `obj` or `schema` must be specified")

        if temp:
            raise com.UnsupportedOperationError(
                "Creating temp tables is not supported by Exasol."
            )

        if database is not None and database != self.current_database:
            raise com.UnsupportedOperationError(
                "Creating tables in other databases is not supported by Exasol"
            )
        else:
            database = None

        quoted = self.compiler.quoted

        if obj is not None:
            if not isinstance(obj, ir.Expr):
                table = ibis.memtable(obj)
            else:
                table = obj

            self._run_pre_execute_hooks(table)

            query = self._to_sqlglot(table)
        else:
            query = None

        type_mapper = self.compiler.type_mapper
        column_defs = [
            sge.ColumnDef(
                this=sg.to_identifier(colname, quoted=quoted),
                kind=type_mapper.from_ibis(typ),
                constraints=(
                    None
                    if typ.nullable
                    else [sge.ColumnConstraint(kind=sge.NotNullColumnConstraint())]
                ),
            )
            for colname, typ in (schema or table.schema()).items()
        ]

        if overwrite:
            temp_name = util.gen_name(f"{self.name}_table")
        else:
            temp_name = name

        table = sg.table(temp_name, catalog=database, quoted=quoted)
        target = sge.Schema(this=table, expressions=column_defs)

        create_stmt = sge.Create(kind="TABLE", this=target)

        this = sg.table(name, catalog=database, quoted=quoted)
        with self._safe_raw_sql(create_stmt):
            if query is not None:
                self.con.execute(
                    sge.Insert(this=table, expression=query).sql(self.name)
                )

            if overwrite:
                self.con.execute(
                    sge.Drop(kind="TABLE", this=this, exists=True).sql(self.name)
                )
                self.con.execute(
                    f"RENAME TABLE {table.sql(self.name)} TO {this.sql(self.name)}"
                )

        if schema is None:
            return self.table(name, database=database)

        # preserve the input schema if it was provided
        return ops.DatabaseTable(
            name, schema=schema, source=self, namespace=ops.Namespace(database=database)
        ).to_expr()

    @property
    def current_schema(self) -> str:
        with self._safe_raw_sql("SELECT CURRENT_SCHEMA") as cur:
            [(schema,)] = cur.fetchall()
        return schema

    def drop_schema(
        self, name: str, database: str | None = None, force: bool = False
    ) -> None:
        if database is not None:
            raise NotImplementedError(
                "`database` argument is not supported for the Exasol backend"
            )
        drop_schema = sg.exp.Drop(kind="SCHEMA", this=name, exists=force)
        with self.begin() as con:
            con.execute(drop_schema.sql(dialect=self.dialect))

    def create_schema(
        self, name: str, database: str | None = None, force: bool = False
    ) -> None:
        if database is not None:
            raise NotImplementedError(
                "`database` argument is not supported for the Exasol backend"
            )
        create_schema = sg.exp.Create(kind="SCHEMA", this=name, exists=force)
        open_schema = self.current_schema
        with self.begin() as con:
            con.execute(create_schema.sql(dialect=self.dialect))
            # Exasol implicitly opens the created schema, therefore we need to restore
            # the previous context.
            con.execute(
                f"OPEN SCHEMA {open_schema}" if open_schema else f"CLOSE SCHEMA {name}"
            )

    def list_schemas(
        self, like: str | None = None, database: str | None = None
    ) -> list[str]:
        if database is not None:
            raise NotImplementedError(
                "`database` argument is not supported for the Exasol backend"
            )

        query = sg.select("schema_name").from_(sg.table("EXA_SCHEMAS", catalog="SYS"))

        with self._safe_raw_sql(query) as con:
            schemas = con.fetchall()
        return self._filter_with_like([schema for (schema,) in schemas], like=like)

    def _cursor_batches(
        self,
        expr: ir.Expr,
        params: Mapping[ir.Scalar, Any] | None = None,
        limit: int | str | None = None,
        chunk_size: int = 1 << 20,
    ) -> Iterable[list]:
        self._run_pre_execute_hooks(expr)

        dtypes = expr.as_table().schema().values()

        with self._safe_raw_sql(
            self.compile(expr, limit=limit, params=params)
        ) as cursor:
            while batch := cursor.fetchmany(chunk_size):
                yield (tuple(map(dt.normalize, dtypes, row)) for row in batch)



from __future__ import annotations

import datetime

from ibis.formats.pandas import PandasData


class ExasolPandasData(PandasData):
    @classmethod
    def convert_String(cls, s, dtype, pandas_type):
        if s.dtype != "object":
            return s.map(str)
        else:
            return s

    @classmethod
    def convert_Interval(cls, s, dtype, pandas_dtype):
        def parse_timedelta(value):
            # format is '(+|-)days hour:minute:second.millisecond'
            days, rest = value.split(" ", 1)
            hms, millis = rest.split(".", 1)
            hours, minutes, seconds = hms.split(":")
            return datetime.timedelta(
                days=int(days),
                hours=int(hours),
                minutes=int(minutes),
                seconds=int(seconds),
                milliseconds=int(millis),
            )

        if s.dtype == "int64":
            # exasol can return intervals as the number of integer days (e.g.,
            # from subtraction of two dates)
            #
            # TODO: investigate whether days are the only interval ever
            # returned as integers
            return s.map(lambda days: datetime.timedelta(days=days))
        return s.map(parse_timedelta, na_action="ignore")



from __future__ import annotations

import datetime
import operator

import pytest
from pytest import param

import ibis
import ibis.expr.datatypes as dt
import ibis.expr.operations as ops
import ibis.expr.types as ir
from ibis import _
from ibis.common.temporal import IntervalUnit
from ibis.expr import api
from ibis.tests.util import assert_equal


def test_temporal_literals():
    date = ibis.literal("2015-01-01", "date")
    assert isinstance(date, ir.DateScalar)

    timestamp = ibis.literal("2017-01-01 12:00:33", "timestamp")
    assert isinstance(timestamp, ir.TimestampScalar)


def test_interval_function_integers():
    # No args, default to 0 seconds
    assert_equal(ibis.interval(), ibis.interval(0, "s"))

    # Default unit is seconds
    assert_equal(ibis.interval(1), ibis.interval(1, "s"))

    # unit is used if provided
    res = ibis.interval(1, "D")
    sol = ibis.literal(1, type=dt.Interval("D"))
    assert_equal(res, sol)


def test_interval_function_timedelta():
    res = ibis.interval(datetime.timedelta())
    sol = ibis.interval()
    assert_equal(res, sol)

    res = ibis.interval(datetime.timedelta(microseconds=10))
    sol = ibis.interval(10, "us")
    assert_equal(res, sol)

    res = ibis.interval(datetime.timedelta(days=5))
    sol = ibis.interval(5, "D")
    assert_equal(res, sol)

    res = ibis.interval(datetime.timedelta(seconds=10, microseconds=2))
    sol = ibis.interval(2, "us") + ibis.interval(10, "s")
    assert_equal(res, sol)


@pytest.mark.parametrize(
    "kw,unit",
    [
        ("nanoseconds", "ns"),
        ("microseconds", "us"),
        ("milliseconds", "ms"),
        ("seconds", "s"),
        ("minutes", "m"),
        ("hours", "h"),
        ("days", "D"),
        ("weeks", "W"),
        ("months", "M"),
        ("quarters", "Q"),
        ("years", "Y"),
    ],
)
def test_interval_function_unit_keywords(kw, unit):
    res = ibis.interval(**{kw: 1})
    sol = ibis.interval(1, unit)
    assert_equal(res, sol)


def test_interval_function_multiple_keywords():
    res = ibis.interval(microseconds=10, hours=3, days=2)
    sol = ibis.interval(10, "us") + ibis.interval(3, "h") + ibis.interval(2, "D")
    assert_equal(res, sol)


def test_interval_function_invalid():
    with pytest.raises(TypeError, match="integer or timedelta"):
        ibis.interval(1.5)

    with pytest.raises(TypeError, match="'value' and 'microseconds'"):
        ibis.interval(1, microseconds=2)


@pytest.mark.parametrize(
    ("interval", "unit", "expected"),
    [
        (api.interval(months=3), "Q", api.interval(quarters=1)),
        (api.interval(months=12), "Y", api.interval(years=1)),
        (api.interval(quarters=8), "Y", api.interval(years=2)),
        (api.interval(days=14), "W", api.interval(weeks=2)),
        (api.interval(minutes=240), "h", api.interval(hours=4)),
        (api.interval(seconds=360), "m", api.interval(minutes=6)),
        (api.interval(milliseconds=5000), "s", api.interval(seconds=5)),
        (api.interval(microseconds=5000000), "s", api.interval(seconds=5)),
        (api.interval(nanoseconds=5000000000), "s", api.interval(seconds=5)),
        (api.interval(seconds=3 * 86400), "D", api.interval(days=3)),
    ],
)
def test_upconvert_interval(interval, unit, expected):
    result = interval.to_unit(unit)
    assert result.equals(expected)


@pytest.mark.parametrize("target", ["Y", "Q", "M"])
@pytest.mark.parametrize(
    "delta",
    [
        api.interval(weeks=1),
        api.interval(days=1),
        api.interval(hours=1),
        api.interval(minutes=1),
        api.interval(seconds=1),
        api.interval(milliseconds=1),
        api.interval(microseconds=1),
        api.interval(nanoseconds=1),
    ],
)
def test_cannot_upconvert(delta, target):
    with pytest.raises(ValueError):
        delta.to_unit(target)


@pytest.mark.parametrize(
    "expr",
    [
        api.interval(days=2) * 2,
        api.interval(days=2) * (-2),
        3 * api.interval(days=2),
        (-3) * api.interval(days=2),
    ],
)
def test_multiply(expr):
    assert isinstance(expr, ir.IntervalScalar)
    assert expr.type().unit == IntervalUnit("D")


@pytest.mark.parametrize(
    ("expr", "expected_unit"),
    [
        (
            api.interval(days=1) + api.interval(days=1),
            IntervalUnit("D"),
        ),
        (
            api.interval(days=2) + api.interval(hours=4),
            IntervalUnit("h"),
        ),
        (
            api.interval(seconds=1) + ibis.interval(minutes=2),
            IntervalUnit("s"),
        ),
    ],
)
def test_add(expr, expected_unit):
    assert isinstance(expr, ir.IntervalScalar)
    assert expr.type().unit == expected_unit


@pytest.mark.parametrize(
    ("expr", "expected_unit"),
    [
        (
            api.interval(days=3) - api.interval(days=1),
            IntervalUnit("D"),
        ),
        (
            api.interval(days=2) - api.interval(hours=4),
            IntervalUnit("h"),
        ),
        (
            api.interval(minutes=2) - api.interval(seconds=1),
            IntervalUnit("s"),
        ),
    ],
)
def test_subtract(expr, expected_unit):
    assert isinstance(expr, ir.IntervalScalar)
    assert expr.type().unit == expected_unit


@pytest.mark.parametrize(
    ("case", "expected"),
    [
        (api.interval(seconds=2).to_unit("s"), api.interval(seconds=2)),
        (
            api.interval(seconds=2).to_unit("ms"),
            api.interval(milliseconds=2 * 1000),
        ),
        (
            api.interval(seconds=2).to_unit("us"),
            api.interval(microseconds=2 * 1000000),
        ),
        (
            api.interval(seconds=2).to_unit("ns"),
            api.interval(nanoseconds=2 * 1000000000),
        ),
        (
            api.interval(milliseconds=2).to_unit("ms"),
            api.interval(milliseconds=2),
        ),
        (
            api.interval(milliseconds=2).to_unit("us"),
            api.interval(microseconds=2 * 1000),
        ),
        (
            api.interval(milliseconds=2).to_unit("ns"),
            api.interval(nanoseconds=2 * 1000000),
        ),
        (
            api.interval(microseconds=2).to_unit("us"),
            api.interval(microseconds=2),
        ),
        (
            api.interval(microseconds=2).to_unit("ns"),
            api.interval(nanoseconds=2 * 1000),
        ),
        (
            api.interval(nanoseconds=2).to_unit("ns"),
            api.interval(nanoseconds=2),
        ),
    ],
)
def test_downconvert_second_parts(case, expected):
    assert isinstance(case, ir.IntervalScalar)
    assert isinstance(expected, ir.IntervalScalar)
    assert case.type().unit == expected.type().unit


@pytest.mark.parametrize(
    ("case", "expected"),
    [
        (api.interval(hours=2).to_unit("h"), api.interval(hours=2)),
        (api.interval(hours=2).to_unit("m"), api.interval(minutes=2 * 60)),
        (api.interval(hours=2).to_unit("s"), api.interval(seconds=2 * 3600)),
        (
            api.interval(hours=2).to_unit("ms"),
            api.interval(milliseconds=2 * 3600000),
        ),
        (
            api.interval(hours=2).to_unit("us"),
            api.interval(microseconds=2 * 3600000000),
        ),
        (
            api.interval(hours=2).to_unit("ns"),
            api.interval(nanoseconds=2 * 3600000000000),
        ),
    ],
)
def test_downconvert_hours(case, expected):
    assert isinstance(case, ir.IntervalScalar)
    assert isinstance(expected, ir.IntervalScalar)
    assert case.type().unit == expected.type().unit


@pytest.mark.parametrize(
    ("case", "expected"),
    [
        (api.interval(minutes=2).to_unit("m"), api.interval(minutes=2)),
        (api.interval(minutes=2).to_unit("s"), api.interval(seconds=2 * 60)),
        (
            api.interval(minutes=2).to_unit("ms"),
            api.interval(milliseconds=2 * 60 * 1000),
        ),
        (
            api.interval(minutes=2).to_unit("us"),
            api.interval(microseconds=2 * 60 * 1000000),
        ),
        (
            api.interval(minutes=2).to_unit("ns"),
            api.interval(nanoseconds=2 * 60 * 1000000000),
        ),
        (api.interval(hours=2).to_unit("h"), api.interval(hours=2)),
        (api.interval(hours=2).to_unit("m"), api.interval(minutes=2 * 60)),
        (api.interval(hours=2).to_unit("s"), api.interval(seconds=2 * 3600)),
        (api.interval(weeks=2).to_unit("D"), api.interval(days=2 * 7)),
        (api.interval(days=2).to_unit("D"), api.interval(days=2)),
        (api.interval(years=2).to_unit("Y"), api.interval(years=2)),
        (api.interval(years=2).to_unit("M"), api.interval(months=24)),
        (api.interval(years=2).to_unit("Q"), api.interval(quarters=8)),
        (api.interval(months=2).to_unit("M"), api.interval(months=2)),
        (api.interval(weeks=2).to_unit("h"), api.interval(hours=2 * 7 * 24)),
        (api.interval(days=2).to_unit("h"), api.interval(hours=2 * 24)),
        (api.interval(days=2).to_unit("m"), api.interval(minutes=2 * 24 * 60)),
        (api.interval(days=2).to_unit("s"), api.interval(seconds=2 * 24 * 60 * 60)),
        (
            api.interval(days=2).to_unit("ms"),
            api.interval(milliseconds=2 * 24 * 60 * 60 * 1_000),
        ),
        (
            api.interval(days=2).to_unit("us"),
            api.interval(microseconds=2 * 24 * 60 * 60 * 1_000_000),
        ),
        (
            api.interval(days=2).to_unit("ns"),
            api.interval(nanoseconds=2 * 24 * 60 * 60 * 1_000_000_000),
        ),
    ],
)
def test_downconvert_interval(case, expected):
    assert isinstance(case, ir.IntervalScalar)
    assert isinstance(expected, ir.IntervalScalar)
    assert case.type().unit == expected.type().unit


@pytest.mark.parametrize(
    ("a", "b", "unit"),
    [
        (api.interval(days=1), api.interval(days=3), "D"),
        (api.interval(seconds=1), api.interval(hours=10), "s"),
        (api.interval(hours=3), api.interval(days=2), "h"),
    ],
)
def test_combine_with_different_kinds(a, b, unit):
    assert (a + b).type().unit == IntervalUnit(unit)


@pytest.mark.parametrize(
    ("case", "expected"),
    [
        (api.interval(quarters=2), api.interval(quarters=2)),
        (api.interval(weeks=2), api.interval(weeks=2)),
        (api.interval(days=3), api.interval(days=3)),
        (api.interval(hours=4), api.interval(hours=4)),
        (api.interval(minutes=5), api.interval(minutes=5)),
        (api.interval(seconds=6), api.interval(seconds=6)),
        (api.interval(milliseconds=7), api.interval(milliseconds=7)),
        (api.interval(microseconds=8), api.interval(microseconds=8)),
        (api.interval(nanoseconds=9), api.interval(nanoseconds=9)),
    ],
)
def test_timedelta_generic_api(case, expected):
    assert case.equals(expected)


ONE_DAY = ibis.interval(days=1)
ONE_HOUR = ibis.interval(hours=1)
ONE_SECOND = ibis.interval(seconds=1)


@pytest.mark.parametrize(
    ("func", "expected_op", "expected_type", "left", "right"),
    [
        # time
        param(
            operator.add,
            ops.TimeAdd,
            ir.TimeColumn,
            lambda _: ONE_HOUR,
            lambda t: t.k,
            id="time_add_interval_column",
        ),
        param(
            operator.add,
            ops.TimeAdd,
            ir.TimeColumn,
            lambda t: t.k,
            lambda _: ONE_HOUR,
            id="time_add_column_interval",
        ),
        param(
            operator.sub,
            ops.TimeSub,
            ir.TimeColumn,
            lambda _: ONE_HOUR,
            lambda t: t.k,
            id="time_sub_interval_column",
            marks=pytest.mark.xfail(
                raises=TypeError,
                reason="interval - time is not yet implemented",
            ),
        ),
        param(
            operator.sub,
            ops.TimeSub,
            ir.TimeColumn,
            lambda t: t.k,
            lambda _: ONE_HOUR,
            id="time_sub_column_interval",
        ),
        # date
        param(
            operator.add,
            ops.DateAdd,
            ir.DateColumn,
            lambda _: ONE_DAY,
            lambda t: t.j,
            id="date_add_interval_column",
        ),
        param(
            operator.add,
            ops.DateAdd,
            ir.DateColumn,
            lambda t: t.j,
            lambda _: ONE_DAY,
            id="date_add_column_interval",
        ),
        param(
            operator.sub,
            ops.DateSub,
            ir.DateColumn,
            lambda _: ONE_DAY,
            lambda t: t.j,
            id="date_sub_interval_column",
            marks=pytest.mark.xfail(
                raises=TypeError,
                reason="interval - date is not yet implemented",
            ),
        ),
        param(
            operator.sub,
            ops.DateSub,
            ir.DateColumn,
            lambda t: t.j,
            lambda _: ONE_DAY,
            id="date_sub_column_interval",
        ),
        # timestamp
        param(
            operator.add,
            ops.TimestampAdd,
            ir.TimestampColumn,
            lambda _: ONE_SECOND,
            lambda t: t.i,
            id="timestamp_add_interval_column",
        ),
        param(
            operator.add,
            ops.TimestampAdd,
            ir.TimestampColumn,
            lambda t: t.i,
            lambda _: ONE_SECOND,
            id="timestamp_add_column_interval",
        ),
        param(
            operator.sub,
            ops.TimestampSub,
            ir.TimestampColumn,
            lambda _: ONE_SECOND,
            lambda t: t.i,
            id="timestamp_sub_interval_column",
            marks=pytest.mark.xfail(
                raises=TypeError,
                reason="interval - timestamp is not yet implemented",
            ),
        ),
        param(
            operator.sub,
            ops.TimestampSub,
            ir.TimestampColumn,
            lambda t: t.i,
            lambda _: ONE_SECOND,
            id="timestamp_sub_column_interval",
        ),
    ],
)
def test_interval_arithmetic(
    table,
    func,
    expected_op,
    expected_type,
    left,
    right,
):
    lhs = left(table)
    rhs = right(table)
    expr = func(lhs, rhs)
    assert isinstance(expr, expected_type)
    assert isinstance(expr.op(), expected_op)


@pytest.mark.parametrize(
    "literal",
    [
        api.interval(3600),
        api.interval(datetime.timedelta(days=3)),
        api.interval(years=1),
        api.interval(quarters=3),
        api.interval(months=2),
        api.interval(weeks=3),
        api.interval(days=-1),
        api.interval(hours=-3),
        api.interval(minutes=5),
        api.interval(seconds=-8),
    ],
)
def test_interval(literal):
    assert isinstance(literal, ir.IntervalScalar)
    repr(literal)  # repr() must return in a reasonable amount of time


def test_timestamp_arithmetics():
    ts1 = api.timestamp(datetime.datetime.now())
    ts2 = api.timestamp(datetime.datetime.today())

    i1 = api.interval(minutes=30)

    # TODO: raise for unsupported operations too
    for expr in [ts2 - ts1, ts1 - ts2]:
        assert isinstance(expr, ir.IntervalScalar)
        assert isinstance(expr.op(), ops.TimestampDiff)
        assert expr.type() == dt.Interval("s")

    for expr in [ts1 - i1, ts2 - i1]:
        assert isinstance(expr, ir.TimestampScalar)
        assert isinstance(expr.op(), ops.TimestampSub)

    for expr in [ts1 + i1, ts2 + i1]:
        assert isinstance(expr, ir.TimestampScalar)
        assert isinstance(expr.op(), ops.TimestampAdd)


def test_date_arithmetics():
    d1 = api.date("2015-01-02")
    d2 = api.date("2017-01-01")
    i1 = api.interval(weeks=3)

    for expr in [d1 - d2, d2 - d1]:
        assert isinstance(expr, ir.IntervalScalar)
        assert isinstance(expr.op(), ops.DateDiff)
        assert expr.type() == dt.Interval("D")

    for expr in [d1 - i1, d2 - i1]:
        assert isinstance(expr, ir.DateScalar)
        assert isinstance(expr.op(), ops.DateSub)

    for expr in [d1 + i1, d2 + i1]:
        assert isinstance(expr, ir.DateScalar)
        assert isinstance(expr.op(), ops.DateAdd)


def test_time_arithmetics():
    t1 = api.time("18:00")
    t2 = api.time("19:12")
    i1 = api.interval(minutes=3)

    for expr in [t1 - t2, t2 - t1]:
        assert isinstance(expr, ir.IntervalScalar)
        assert isinstance(expr.op(), ops.TimeDiff)
        assert expr.type() == dt.Interval("s")

    for expr in [t1 - i1, t2 - i1]:
        assert isinstance(expr, ir.TimeScalar)
        assert isinstance(expr.op(), ops.TimeSub)

    for expr in [t1 + i1, t2 + i1]:
        assert isinstance(expr, ir.TimeScalar)
        assert isinstance(expr.op(), ops.TimeAdd)


def test_invalid_date_arithmetics():
    d1 = api.date("2015-01-02")
    i1 = api.interval(seconds=300)
    i2 = api.interval(minutes=15)
    i3 = api.interval(hours=1)

    for i in [i1, i2, i3]:
        with pytest.raises(TypeError):
            d1 - i
        with pytest.raises(TypeError):
            d1 + i


@pytest.mark.parametrize(
    ("value", "prop", "expected_unit"),
    [
        (api.interval(seconds=3600), "nanoseconds", "ns"),
        (api.interval(seconds=3600), "microseconds", "us"),
        (api.interval(seconds=3600), "milliseconds", "ms"),
        (api.interval(seconds=3600), "seconds", "s"),
        (api.interval(seconds=3600), "minutes", "m"),
        (api.interval(seconds=3600), "hours", "h"),
        (api.interval(weeks=3), "days", "D"),
        (api.interval(days=21), "weeks", "W"),
    ],
)
def test_interval_properties(value, prop, expected_unit):
    assert getattr(value, prop).type().unit == IntervalUnit(expected_unit)


@pytest.mark.parametrize(
    "interval",
    [api.interval(years=1), api.interval(quarters=4), api.interval(months=12)],
)
@pytest.mark.parametrize(
    ("prop", "expected_unit"),
    [("months", "M"), ("quarters", "Q"), ("years", "Y")],
)
def test_interval_monthly_properties(interval, prop, expected_unit):
    assert getattr(interval, prop).type().unit == IntervalUnit(expected_unit)


@pytest.mark.parametrize(
    ("interval", "prop"),
    [
        (api.interval(hours=48), "months"),
        (api.interval(years=2), "seconds"),
        (api.interval(quarters=1), "weeks"),
    ],
)
def test_unsupported_properties(interval, prop):
    with pytest.raises(ValueError):
        getattr(interval, prop)


@pytest.mark.parametrize("column", ["a", "b", "c", "d"])  # integer columns
@pytest.mark.parametrize(
    "unit", ["Y", "Q", "M", "D", "W", "h", "m", "s", "ms", "us", "ns"]
)
def test_integer_to_interval(column, unit, table):
    c = table[column]
    i = c.to_interval(unit)
    assert isinstance(i, ir.IntervalColumn)
    assert i.type().unit == IntervalUnit(unit)


@pytest.mark.parametrize(
    "unit", ["Y", "Q", "M", "D", "W", "h", "m", "s", "ms", "us", "ns"]
)
@pytest.mark.parametrize(
    "operands",
    [
        lambda t, u: (api.interval(3, unit=u), api.interval(2, unit=u)),
        lambda t, u: (api.interval(3, unit=u), api.interval(3, unit=u)),
        lambda t, u: (t.c.to_interval(unit=u), api.interval(2, unit=u)),
        lambda t, u: (t.c.to_interval(unit=u), t.d.to_interval(unit=u)),
    ],
)
@pytest.mark.parametrize(
    "operator",
    [
        operator.eq,
        operator.ne,
        operator.ge,
        operator.gt,
        operator.le,
        operator.lt,
    ],
    ids=lambda op: op.__name__,
)
def test_interval_comparisons(unit, operands, operator, table):
    a, b = operands(table, unit)
    expr = operator(a, b)

    assert isinstance(a, ir.IntervalValue)
    assert isinstance(b, ir.IntervalValue)
    assert isinstance(expr, ir.BooleanValue)


@pytest.mark.parametrize(
    "operands",
    [
        lambda t: (api.date("2016-01-01"), api.date("2016-02-02")),
        lambda t: (t.j, api.date("2016-01-01")),
        lambda t: (api.date("2016-01-01"), t.j),
        lambda t: (t.j, t.i.date()),
        lambda t: (t.i.date(), t.j),
    ],
    ids=[
        "literal-literal",
        "column-literal",
        "literal-column",
        "column-casted",
        "casted-column",
    ],
)
@pytest.mark.parametrize(
    "interval",
    [
        lambda t: api.interval(years=4),
        lambda t: api.interval(quarters=4),
        lambda t: api.interval(months=3),
        lambda t: api.interval(weeks=2),
        lambda t: api.interval(days=1),
        lambda t: t.c.to_interval(unit="Y"),
        lambda t: t.c.to_interval(unit="M"),
        lambda t: t.c.to_interval(unit="W"),
        lambda t: t.c.to_interval(unit="D"),
    ],
    ids=[
        "years",
        "quarters",
        "months",
        "weeks",
        "days",
        "to-years",
        "to-months",
        "to-weeks",
        "to-days",
    ],
)
@pytest.mark.parametrize(
    "arithmetic",
    [lambda a, i: a - i, lambda a, i: a + i, lambda a, i: i + a],
    ids=["subtract", "radd", "add"],
)
@pytest.mark.parametrize(
    "operator",
    [
        operator.eq,
        operator.ne,
        operator.ge,
        operator.gt,
        operator.le,
        operator.lt,
    ],
    ids=lambda op: op.__name__,
)
def test_complex_date_comparisons(operands, interval, arithmetic, operator, table):
    (a, b), i = operands(table), interval(table)

    a_ = arithmetic(a, i)
    expr = operator(a_, b)

    assert isinstance(a, ir.DateValue)
    assert isinstance(b, ir.DateValue)
    assert isinstance(a_, ir.DateValue)
    assert isinstance(i, ir.IntervalValue)
    assert isinstance(expr, ir.BooleanValue)


def test_interval_column_name(table):
    c = table.i
    expr = (c - c).name("foo")
    assert expr.get_name() == "foo"


@pytest.mark.parametrize(
    "operand",
    [lambda t: api.timestamp(datetime.datetime.now()), lambda t: t.i],
    ids=["column", "literal"],
)
@pytest.mark.parametrize(
    "unit",
    [
        "Y",
        "y",
        "year",
        "YEAR",
        "YYYY",
        "SYYYY",
        "YYY",
        "YY",
        "Q",
        "q",
        "quarter",
        "QUARTER",
        "M",
        "month",
        "MONTH",
        "w",
        "W",
        "week",
        "WEEK",
        "d",
        "J",
        "day",
        "DAY",
        "h",
        "H",
        "HH24",
        "hour",
        "HOUR",
        "m",
        "MI",
        "minute",
        "MINUTE",
        "s",
        "second",
        "SECOND",
        "ms",
        "millisecond",
        "MILLISECOND",
        "us",
        "microsecond",
        "MICROSECOND",
        "ns",
        "nanosecond",
        "NANOSECOND",
    ],
)
def test_timestamp_truncate(table, operand, unit):
    expr = operand(table).truncate(unit)
    assert isinstance(expr, ir.TimestampValue)
    assert isinstance(expr.op(), ops.TimestampTruncate)


@pytest.mark.parametrize("operand", [lambda t: api.date("2018-01-01"), lambda t: t.j])
@pytest.mark.parametrize("unit", ["Y", "Q", "M", "D", "W"])
def test_date_truncate(table, operand, unit):
    expr = operand(table).truncate(unit)
    assert isinstance(expr, ir.DateValue)
    assert isinstance(expr.op(), ops.DateTruncate)


@pytest.mark.parametrize("operand", [lambda t: api.time("18:00"), lambda t: t.k])
@pytest.mark.parametrize("unit", ["h", "m", "s", "ms", "us", "ns"])
def test_time_truncate(table, operand, unit):
    expr = operand(table).truncate(unit)
    assert isinstance(expr, ir.TimeValue)
    assert isinstance(expr.op(), ops.TimeTruncate)


def test_date_literal():
    expr = ibis.date(2022, 2, 4)
    sol = ops.DateFromYMD(2022, 2, 4).to_expr()
    assert expr.equals(sol)

    expr = ibis.date("2022-02-04")
    sol = ibis.literal("2022-02-04", type=dt.date)
    assert expr.equals(sol)


def test_date_expression():
    t = ibis.table({"x": "int", "y": "int", "z": "int", "s": "string"})
    deferred = ibis.date(_.x, _.y, _.z)
    expr = ibis.date(t.x, t.y, t.z)
    assert isinstance(expr.op(), ops.DateFromYMD)
    assert deferred.resolve(t).equals(expr)
    assert repr(deferred) == "date(_.x, _.y, _.z)"

    deferred = ibis.date(_.s)
    expr = ibis.date(t.s)
    assert deferred.resolve(t).equals(expr)
    assert repr(deferred) == "date(_.s)"


def test_time_literal():
    expr = ibis.time(1, 2, 3)
    sol = ops.TimeFromHMS(1, 2, 3).to_expr()
    assert expr.equals(sol)

    expr = ibis.time("01:02:03")
    sol = ibis.literal("01:02:03", type=dt.time)
    assert expr.equals(sol)


def test_time_expression():
    t = ibis.table({"x": "int", "y": "int", "z": "int", "s": "string"})
    deferred = ibis.time(_.x, _.y, _.z)
    expr = ibis.time(t.x, t.y, t.z)
    assert isinstance(expr.op(), ops.TimeFromHMS)
    assert deferred.resolve(t).equals(expr)
    assert repr(deferred) == "time(_.x, _.y, _.z)"

    deferred = ibis.time(_.s)
    expr = ibis.time(t.s)
    assert deferred.resolve(t).equals(expr)
    assert repr(deferred) == "time(_.s)"


def test_timestamp_literals():
    assert ibis.timestamp(2022, 2, 4, 16, 20, 00).type() == dt.timestamp


def test_timestamp_literal():
    expr = ibis.timestamp(2022, 2, 4, 16, 20, 0)
    sol = ops.TimestampFromYMDHMS(2022, 2, 4, 16, 20, 0).to_expr()
    assert expr.equals(sol)

    expr = ibis.timestamp("2022-02-04T01:02:03")
    sol = ibis.literal("2022-02-04T01:02:03", type=dt.timestamp)
    assert expr.equals(sol)

    expr = ibis.timestamp("2022-02-04T01:02:03Z")
    sol = ibis.literal("2022-02-04T01:02:03", type=dt.Timestamp(timezone="UTC"))
    assert expr.equals(sol)


def test_timestamp_expression():
    t = ibis.table(dict.fromkeys("abcdef", "int"))
    deferred = ibis.timestamp(_.a, _.b, _.c, _.d, _.e, _.f)
    expr = ibis.timestamp(t.a, t.b, t.c, t.d, t.e, t.f)
    assert isinstance(expr.op(), ops.TimestampFromYMDHMS)
    assert deferred.resolve(t).equals(expr)
    assert repr(deferred) == "timestamp(_.a, _.b, _.c, _.d, _.e, _.f)"

    t2 = ibis.table({"s": "string"})
    deferred = ibis.timestamp(_.s, timezone="UTC")
    expr = ibis.timestamp(t2.s, timezone="UTC")
    assert deferred.resolve(t2).equals(expr)
    assert repr(deferred) == "timestamp(_.s, timezone='UTC')"


def test_timestamp_bucket():
    ts = ibis.table({"ts": "timestamp"}).ts

    components = [
        "nanoseconds",
        "microseconds",
        "milliseconds",
        "seconds",
        "minutes",
        "hours",
        "days",
        "weeks",
        "months",
        "quarters",
        "years",
    ]
    for component in components:
        kws = {component: 2}
        expr1 = ts.bucket(**kws)
        expr2 = ts.bucket(ibis.interval(**kws))
        assert expr1.equals(expr2)

    with pytest.raises(
        ValueError, match="Must specify either interval value or components"
    ):
        ts.bucket(ibis.interval(seconds=1), minutes=2)

    with pytest.raises(
        ValueError, match="Must specify either interval value or components"
    ):
        ts.bucket()



"""Tests for JSON operations."""
from __future__ import annotations

import sqlite3

import pandas as pd
import pandas.testing as tm
import pytest
from packaging.version import parse as vparse
from pytest import param

pytestmark = [
    pytest.mark.never(["impala"], reason="doesn't support JSON and never will"),
    pytest.mark.notyet(["clickhouse"], reason="upstream is broken"),
    pytest.mark.notimpl(["datafusion", "exasol", "mssql", "druid", "oracle"]),
]


@pytest.mark.parametrize(
    ("expr_fn", "expected"),
    [
        param(
            lambda t: t.js["a"].name("res"),
            pd.Series([[1, 2, 3, 4], None, "foo"] + [None] * 3, name="res"),
            id="object",
        ),
        param(
            lambda t: t.js[1].name("res"),
            pd.Series([None] * 4 + [47, None], dtype="object", name="res"),
            id="array",
        ),
    ],
)
@pytest.mark.notyet(
    ["sqlite"],
    condition=vparse(sqlite3.sqlite_version) < vparse("3.38.0"),
    reason="JSON not supported in SQLite < 3.38.0",
)
@pytest.mark.broken(
    ["flink"],
    reason="https://github.com/ibis-project/ibis/pull/6920#discussion_r1373212503",
)
@pytest.mark.broken(
    ["risingwave"], reason="TODO(Kexiang): order mismatch in array", strict=False
)
def test_json_getitem(json_t, expr_fn, expected):
    expr = expr_fn(json_t)
    result = expr.execute()
    tm.assert_series_equal(result.fillna(pd.NA), expected.fillna(pd.NA))


@pytest.mark.notimpl(["dask", "mysql", "pandas", "risingwave"])
@pytest.mark.notyet(["bigquery", "sqlite"], reason="doesn't support maps")
@pytest.mark.notyet(["postgres"], reason="only supports map<string, string>")
@pytest.mark.notyet(
    ["pyspark", "trino", "flink"], reason="should work but doesn't deserialize JSON"
)
def test_json_map(backend, json_t):
    expr = json_t.js.map.name("res")
    result = expr.execute()
    expected = pd.Series(
        [
            {"a": [1, 2, 3, 4], "b": 1},
            {"a": None, "b": 2},
            {"a": "foo", "c": None},
            None,
            None,
            None,
        ],
        dtype="object",
        name="res",
    )
    backend.assert_series_equal(result, expected)


@pytest.mark.notimpl(["dask", "mysql", "pandas", "risingwave"])
@pytest.mark.notyet(["sqlite"], reason="doesn't support arrays")
@pytest.mark.notyet(
    ["pyspark", "trino", "flink"], reason="should work but doesn't deserialize JSON"
)
@pytest.mark.notyet(["bigquery"], reason="doesn't allow null in arrays")
def test_json_array(backend, json_t):
    expr = json_t.js.array.name("res")
    result = expr.execute()
    expected = pd.Series(
        [None, None, None, None, [42, 47, 55], []], name="res", dtype="object"
    )
    backend.assert_series_equal(result, expected)



from __future__ import annotations

import collections
from keyword import iskeyword
from typing import TYPE_CHECKING

from public import public

import ibis.expr.operations as ops
from ibis.common.deferred import deferrable
from ibis.expr.types.generic import Column, Scalar, Value, literal

if TYPE_CHECKING:
    from collections.abc import Iterable, Mapping, Sequence

    import ibis.expr.datatypes as dt
    import ibis.expr.types as ir
    from ibis.expr.types.typing import V


@public
@deferrable
def struct(
    value: Iterable[tuple[str, V]] | Mapping[str, V],
    type: str | dt.DataType | None = None,
) -> StructValue:
    """Create a struct expression.

    If the input expressions are all column expressions, then the output will be
    a `StructColumn`.

    If the input expressions are Python literals, then the output will be a
    `StructScalar`.

    Parameters
    ----------
    value
        The underlying data for literal struct value or a pairs of field names
        and column expressions.
    type
        An instance of `ibis.expr.datatypes.DataType` or a string indicating
        the ibis type of `value`. This is only used if all of the input values
        are literals.

    Returns
    -------
    StructValue
        An expression representing a literal or column struct (compound type with
        fields of fixed types)

    Examples
    --------
    Create a struct literal from a [](`dict`) with the type inferred
    >>> import ibis
    >>> t = ibis.struct(dict(a=1, b="foo"))

    Create a struct literal from a [](`dict`) with a specified type
    >>> t = ibis.struct(dict(a=1, b="foo"), type="struct<a: float, b: string>")

    Specify a specific type for the struct literal
    >>> t = ibis.struct(dict(a=1, b=40), type="struct<a: float, b: int32>")

    Create a struct array from multiple arrays
    >>> ibis.options.interactive = True
    >>> t = ibis.memtable({"a": [1, 2, 3], "b": ["foo", "bar", "baz"]})
    >>> ibis.struct([("a", t.a), ("b", t.b)])
    ┏━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┓
    ┃ StructColumn()              ┃
    ┡━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┩
    │ struct<a: int64, b: string> │
    ├─────────────────────────────┤
    │ {'a': 1, 'b': 'foo'}        │
    │ {'a': 2, 'b': 'bar'}        │
    │ {'a': 3, 'b': 'baz'}        │
    └─────────────────────────────┘

    Create a struct array from columns and literals
    >>> ibis.struct([("a", t.a), ("b", "foo")])
    ┏━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┓
    ┃ StructColumn()              ┃
    ┡━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┩
    │ struct<a: int64, b: string> │
    ├─────────────────────────────┤
    │ {'a': 1, 'b': 'foo'}        │
    │ {'a': 2, 'b': 'foo'}        │
    │ {'a': 3, 'b': 'foo'}        │
    └─────────────────────────────┘
    """
    import ibis.expr.operations as ops

    fields = dict(value)
    if any(isinstance(value, Value) for value in fields.values()):
        names = tuple(fields.keys())
        values = tuple(fields.values())
        return ops.StructColumn(names=names, values=values).to_expr()
    else:
        return literal(collections.OrderedDict(fields), type=type)


@public
class StructValue(Value):
    """A struct literal or column.

    Can be constructed with [`ibis.struct()`](#ibis.expr.types.struct).

    Examples
    --------
    >>> import ibis
    >>> ibis.options.interactive = True
    >>> t = ibis.memtable({"s": [{"a": 1, "b": "foo"}, {"a": 3, "b": None}, None]})
    >>> t
    ┏━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┓
    ┃ s                           ┃
    ┡━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┩
    │ struct<a: int64, b: string> │
    ├─────────────────────────────┤
    │ {'a': 1, 'b': 'foo'}        │
    │ {'a': 3, 'b': None}         │
    │ NULL                        │
    └─────────────────────────────┘

    Can use either `.` or `[]` to access fields:

    >>> t.s.a
    ┏━━━━━━━┓
    ┃ a     ┃
    ┡━━━━━━━┩
    │ int64 │
    ├───────┤
    │     1 │
    │     3 │
    │  NULL │
    └───────┘
    >>> t.s["a"]
    ┏━━━━━━━┓
    ┃ a     ┃
    ┡━━━━━━━┩
    │ int64 │
    ├───────┤
    │     1 │
    │     3 │
    │  NULL │
    └───────┘
    """

    def __dir__(self):
        out = set(dir(type(self)))
        out.update(
            c for c in self.type().names if c.isidentifier() and not iskeyword(c)
        )
        return sorted(out)

    def _ipython_key_completions_(self) -> list[str]:
        return sorted(self.type().names)

    def __getitem__(self, name: str) -> ir.Value:
        """Extract the `name` field from this struct.

        Parameters
        ----------
        name
            The name of the field to access.

        Returns
        -------
        Value
            An expression with the type of the field being accessed.

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> t = ibis.memtable({"s": [{"a": 1, "b": "foo"}, {"a": 3, "b": None}, None]})
        >>> t
        ┏━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┓
        ┃ s                           ┃
        ┡━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┩
        │ struct<a: int64, b: string> │
        ├─────────────────────────────┤
        │ {'a': 1, 'b': 'foo'}        │
        │ {'a': 3, 'b': None}         │
        │ NULL                        │
        └─────────────────────────────┘
        >>> t.s["a"]
        ┏━━━━━━━┓
        ┃ a     ┃
        ┡━━━━━━━┩
        │ int64 │
        ├───────┤
        │     1 │
        │     3 │
        │  NULL │
        └───────┘
        >>> t.s["b"]
        ┏━━━━━━━━┓
        ┃ b      ┃
        ┡━━━━━━━━┩
        │ string │
        ├────────┤
        │ foo    │
        │ NULL   │
        │ NULL   │
        └────────┘
        >>> t.s["foo_bar"]
        Traceback (most recent call last):
            ...
        KeyError: 'foo_bar'
        """
        if name not in self.names:
            raise KeyError(name)
        return ops.StructField(self, name).to_expr()

    def __setstate__(self, instance_dictionary):
        self.__dict__ = instance_dictionary

    def __getattr__(self, name: str) -> ir.Value:
        """Extract the `name` field from this struct.

        Parameters
        ----------
        name
            The name of the field to access.

        Returns
        -------
        Value
            An expression with the type of the field being accessed.

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> t = ibis.memtable({"s": [{"a": 1, "b": "foo"}, {"a": 3, "b": None}, None]})
        >>> t
        ┏━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┓
        ┃ s                           ┃
        ┡━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┩
        │ struct<a: int64, b: string> │
        ├─────────────────────────────┤
        │ {'a': 1, 'b': 'foo'}        │
        │ {'a': 3, 'b': None}         │
        │ NULL                        │
        └─────────────────────────────┘
        >>> t.s.a
        ┏━━━━━━━┓
        ┃ a     ┃
        ┡━━━━━━━┩
        │ int64 │
        ├───────┤
        │     1 │
        │     3 │
        │  NULL │
        └───────┘
        >>> t.s.b
        ┏━━━━━━━━┓
        ┃ b      ┃
        ┡━━━━━━━━┩
        │ string │
        ├────────┤
        │ foo    │
        │ NULL   │
        │ NULL   │
        └────────┘
        >>> t.s.foo_bar
        Traceback (most recent call last):
            ...
        AttributeError: foo_bar
        """
        try:
            return self[name]
        except KeyError:
            raise AttributeError(name) from None

    @property
    def names(self) -> Sequence[str]:
        """Return the field names of the struct."""
        return self.type().names

    @property
    def types(self) -> Sequence[dt.DataType]:
        """Return the field types of the struct."""
        return self.type().types

    @property
    def fields(self) -> Mapping[str, dt.DataType]:
        """Return a mapping from field name to field type of the struct."""
        return self.type().fields

    def lift(self) -> ir.Table:
        """Project the fields of `self` into a table.

        This method is useful when analyzing data that has deeply nested
        structs or arrays of structs. `lift` can be chained to avoid repeating
        column names and table references.

        Returns
        -------
        Table
            A projection with this struct expression's fields.

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> t = ibis.memtable(
        ...     {
        ...         "pos": [
        ...             {"lat": 10.1, "lon": 30.3},
        ...             {"lat": 10.2, "lon": 30.2},
        ...             {"lat": 10.3, "lon": 30.1},
        ...         ]
        ...     }
        ... )
        >>> t
        ┏━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┓
        ┃ pos                                ┃
        ┡━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┩
        │ struct<lat: float64, lon: float64> │
        ├────────────────────────────────────┤
        │ {'lat': 10.1, 'lon': 30.3}         │
        │ {'lat': 10.2, 'lon': 30.2}         │
        │ {'lat': 10.3, 'lon': 30.1}         │
        └────────────────────────────────────┘
        >>> t.pos.lift()
        ┏━━━━━━━━━┳━━━━━━━━━┓
        ┃ lat     ┃ lon     ┃
        ┡━━━━━━━━━╇━━━━━━━━━┩
        │ float64 │ float64 │
        ├─────────┼─────────┤
        │    10.1 │    30.3 │
        │    10.2 │    30.2 │
        │    10.3 │    30.1 │
        └─────────┴─────────┘

        See Also
        --------
        [`Table.unpack`](./expression-tables.qmd#ibis.expr.types.relations.Table.unpack)
        """
        import ibis.expr.analysis as an

        # TODO(kszucs): avoid expression roundtripping
        table = an.find_first_base_table(self.op()).to_expr()
        return table[[self[name] for name in self.names]]

    def destructure(self) -> list[ir.Value]:
        """Destructure a ``StructValue`` into the corresponding struct fields.

        When assigned, a destruct value will be destructured and assigned to
        multiple columns.

        Returns
        -------
        list[AnyValue]
            Value expressions corresponding to the struct fields.

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> t = ibis.memtable({"s": [{"a": 1, "b": "foo"}, {"a": 3, "b": None}, None]})
        >>> t
        ┏━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┓
        ┃ s                           ┃
        ┡━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┩
        │ struct<a: int64, b: string> │
        ├─────────────────────────────┤
        │ {'a': 1, 'b': 'foo'}        │
        │ {'a': 3, 'b': None}         │
        │ NULL                        │
        └─────────────────────────────┘
        >>> a, b = t.s.destructure()
        >>> a
        ┏━━━━━━━┓
        ┃ a     ┃
        ┡━━━━━━━┩
        │ int64 │
        ├───────┤
        │     1 │
        │     3 │
        │  NULL │
        └───────┘
        >>> b
        ┏━━━━━━━━┓
        ┃ b      ┃
        ┡━━━━━━━━┩
        │ string │
        ├────────┤
        │ foo    │
        │ NULL   │
        │ NULL   │
        └────────┘
        """
        return [self[field_name] for field_name in self.type().names]


@public
class StructScalar(Scalar, StructValue):
    pass


@public
class StructColumn(Column, StructValue):
    def __getitem__(self, name: str) -> ir.Column:
        return StructValue.__getitem__(self, name)



from __future__ import annotations

import ibis

to_sql = ibis.bigquery.compile


@ibis.udf.scalar.builtin
def farm_fingerprint(value: bytes) -> int:
    ...


@ibis.udf.scalar.builtin(schema="fn", database="bqutil")
def from_hex(value: str) -> int:
    """Community function to convert from hex string to integer.

    See:
    https://github.com/GoogleCloudPlatform/bigquery-utils/tree/master/udfs/community#from_hexvalue-string
    """


def test_bqutil_fn_from_hex(snapshot):
    # Project ID should be enclosed in backticks.
    expr = from_hex("face")
    snapshot.assert_match(to_sql(expr), "out.sql")


def test_farm_fingerprint(snapshot):
    # No backticks needed if there is no schema defined.
    expr = farm_fingerprint(b"Hello, World!")
    snapshot.assert_match(to_sql(expr), "out.sql")



from __future__ import annotations

import ibis

from .conftest import tpch_test


@tpch_test
def test_tpc_h11(partsupp, supplier, nation):
    NATION = "GERMANY"
    FRACTION = 0.0001

    q = partsupp
    q = q.join(supplier, partsupp.ps_suppkey == supplier.s_suppkey)
    q = q.join(nation, nation.n_nationkey == supplier.s_nationkey)

    q = q.filter([q.n_name == NATION])

    innerq = partsupp
    innerq = innerq.join(supplier, partsupp.ps_suppkey == supplier.s_suppkey)
    innerq = innerq.join(nation, nation.n_nationkey == supplier.s_nationkey)
    innerq = innerq.filter([innerq.n_name == NATION])
    innerq = innerq.aggregate(total=(innerq.ps_supplycost * innerq.ps_availqty).sum())

    gq = q.group_by([q.ps_partkey])
    q = gq.aggregate(value=(q.ps_supplycost * q.ps_availqty).sum())
    q = q.filter([q.value > innerq.total * FRACTION])
    q = q.order_by(ibis.desc(q.value))
    return q



from __future__ import annotations

import abc
from typing import TYPE_CHECKING, Any, ClassVar

import sqlglot as sg
import sqlglot.expressions as sge

import ibis
import ibis.expr.operations as ops
import ibis.expr.schema as sch
import ibis.expr.types as ir
from ibis import util
from ibis.backends.base import BaseBackend
from ibis.backends.base.sqlglot.compiler import STAR

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator, Mapping

    import pandas as pd
    import pyarrow as pa

    import ibis.expr.datatypes as dt
    from ibis.backends.base.sqlglot.compiler import SQLGlotCompiler
    from ibis.common.typing import SupportsSchema


class SQLGlotBackend(BaseBackend):
    compiler: ClassVar[SQLGlotCompiler]
    name: ClassVar[str]

    @property
    def dialect(self) -> sg.Dialect:
        return self.compiler.dialect

    @classmethod
    def has_operation(cls, operation: type[ops.Value]) -> bool:
        compiler = cls.compiler
        method = getattr(compiler, f"visit_{operation.__name__}", None)
        return method is not None and method not in (
            compiler.visit_Undefined,
            compiler.visit_Unsupported,
        )

    def _fetch_from_cursor(self, cursor, schema: sch.Schema) -> pd.DataFrame:
        import pandas as pd

        from ibis.formats.pandas import PandasData

        try:
            df = pd.DataFrame.from_records(
                cursor, columns=schema.names, coerce_float=True
            )
        except Exception:
            # clean up the cursor if we fail to create the DataFrame
            #
            # in the sqlite case failing to close the cursor results in
            # artificially locked tables
            cursor.close()
            raise
        df = PandasData.convert_table(df, schema)
        return df

    def table(
        self, name: str, schema: str | None = None, database: str | None = None
    ) -> ir.Table:
        """Construct a table expression.

        Parameters
        ----------
        name
            Table name
        schema
            Schema name
        database
            Database name

        Returns
        -------
        Table
            Table expression

        """
        table_schema = self.get_schema(name, schema=schema, database=database)
        return ops.DatabaseTable(
            name,
            schema=table_schema,
            source=self,
            namespace=ops.Namespace(database=database, schema=schema),
        ).to_expr()

    def _to_sqlglot(
        self, expr: ir.Expr, *, limit: str | None = None, params=None, **_: Any
    ):
        """Compile an Ibis expression to a sqlglot object."""
        table_expr = expr.as_table()

        if limit == "default":
            limit = ibis.options.sql.default_limit
        if limit is not None:
            table_expr = table_expr.limit(limit)

        if params is None:
            params = {}

        sql = self.compiler.translate(table_expr.op(), params=params)
        assert not isinstance(sql, sge.Subquery)

        if isinstance(sql, sge.Table):
            sql = sg.select(STAR).from_(sql)

        assert not isinstance(sql, sge.Subquery)
        return sql

    def compile(
        self, expr: ir.Expr, limit: str | None = None, params=None, **kwargs: Any
    ):
        """Compile an Ibis expression to a SQL string."""
        query = self._to_sqlglot(expr, limit=limit, params=params, **kwargs)
        sql = query.sql(dialect=self.dialect, pretty=True)
        self._log(sql)
        return sql

    def _to_sql(self, expr: ir.Expr, **kwargs) -> str:
        return self.compile(expr, **kwargs)

    def _log(self, sql: str) -> None:
        """Log `sql`.

        This method can be implemented by subclasses. Logging occurs when
        `ibis.options.verbose` is `True`.
        """
        from ibis import util

        util.log(sql)

    def sql(
        self,
        query: str,
        schema: SupportsSchema | None = None,
        dialect: str | None = None,
    ) -> ir.Table:
        query = self._transpile_sql(query, dialect=dialect)
        if schema is None:
            schema = self._get_schema_using_query(query)
        return ops.SQLQueryResult(query, ibis.schema(schema), self).to_expr()

    # TODO(kszucs): should be removed in favor of _get_schema_using_query()
    @abc.abstractmethod
    def _metadata(self, query: str) -> Iterator[tuple[str, dt.DataType]]:
        """Return the metadata of a SQL query."""

    def _get_schema_using_query(self, query: str) -> sch.Schema:
        """Return an ibis Schema from a backend-specific SQL string."""
        return sch.Schema.from_tuples(self._metadata(query))

    def _get_sql_string_view_schema(self, name, table, query) -> sch.Schema:
        compiler = self.compiler
        dialect = compiler.dialect

        cte = self._to_sqlglot(table)
        parsed = sg.parse_one(query, read=dialect)
        parsed.args["with"] = cte.args.pop("with", [])
        parsed = parsed.with_(
            sg.to_identifier(name, quoted=compiler.quoted), as_=cte, dialect=dialect
        )

        sql = parsed.sql(dialect)
        return self._get_schema_using_query(sql)

    def create_view(
        self,
        name: str,
        obj: ir.Table,
        *,
        database: str | None = None,
        schema: str | None = None,
        overwrite: bool = False,
    ) -> ir.Table:
        src = sge.Create(
            this=sg.table(
                name, db=schema, catalog=database, quoted=self.compiler.quoted
            ),
            kind="VIEW",
            replace=overwrite,
            expression=self.compile(obj),
        )
        self._register_in_memory_tables(obj)
        with self._safe_raw_sql(src):
            pass
        return self.table(name, database=database)

    def _register_in_memory_tables(self, expr: ir.Expr) -> None:
        for memtable in expr.op().find(ops.InMemoryTable):
            self._register_in_memory_table(memtable)

    def drop_view(
        self,
        name: str,
        *,
        database: str | None = None,
        schema: str | None = None,
        force: bool = False,
    ) -> None:
        src = sge.Drop(
            this=sg.table(
                name, db=schema, catalog=database, quoted=self.compiler.quoted
            ),
            kind="VIEW",
            exists=force,
        )
        with self._safe_raw_sql(src):
            pass

    def _load_into_cache(self, name, expr):
        self.create_table(name, expr, schema=expr.schema(), temp=True)

    def _clean_up_cached_table(self, op):
        self.drop_table(op.name)

    def execute(
        self,
        expr: ir.Expr,
        params: Mapping | None = None,
        limit: str | None = "default",
        **kwargs: Any,
    ) -> Any:
        """Execute an expression."""

        self._run_pre_execute_hooks(expr)
        table = expr.as_table()
        sql = self.compile(table, params=params, limit=limit, **kwargs)

        schema = table.schema()

        # TODO(kszucs): these methods should be abstractmethods or this default
        # implementation should be removed
        with self._safe_raw_sql(sql) as cur:
            result = self._fetch_from_cursor(cur, schema)
        return expr.__pandas_result__(result)

    def drop_table(
        self,
        name: str,
        database: str | None = None,
        schema: str | None = None,
        force: bool = False,
    ) -> None:
        drop_stmt = sg.exp.Drop(
            kind="TABLE",
            this=sg.table(
                name, db=schema, catalog=database, quoted=self.compiler.quoted
            ),
            exists=force,
        )
        with self._safe_raw_sql(drop_stmt):
            pass

    def _cursor_batches(
        self,
        expr: ir.Expr,
        params: Mapping[ir.Scalar, Any] | None = None,
        limit: int | str | None = None,
        chunk_size: int = 1 << 20,
    ) -> Iterable[list]:
        self._run_pre_execute_hooks(expr)

        with self._safe_raw_sql(
            self.compile(expr, limit=limit, params=params)
        ) as cursor:
            while batch := cursor.fetchmany(chunk_size):
                yield batch

    @util.experimental
    def to_pyarrow_batches(
        self,
        expr: ir.Expr,
        *,
        params: Mapping[ir.Scalar, Any] | None = None,
        limit: int | str | None = None,
        chunk_size: int = 1_000_000,
        **_: Any,
    ) -> pa.ipc.RecordBatchReader:
        """Execute expression and return an iterator of pyarrow record batches.

        This method is eager and will execute the associated expression
        immediately.

        Parameters
        ----------
        expr
            Ibis expression to export to pyarrow
        limit
            An integer to effect a specific row limit. A value of `None` means
            "no limit". The default is in `ibis/config.py`.
        params
            Mapping of scalar parameter expressions to value.
        chunk_size
            Maximum number of rows in each returned record batch.

        Returns
        -------
        RecordBatchReader
            Collection of pyarrow `RecordBatch`s.

        """
        pa = self._import_pyarrow()

        schema = expr.as_table().schema()
        array_type = schema.as_struct().to_pyarrow()
        arrays = (
            pa.array(map(tuple, batch), type=array_type)
            for batch in self._cursor_batches(
                expr, params=params, limit=limit, chunk_size=chunk_size
            )
        )
        batches = map(pa.RecordBatch.from_struct_array, arrays)

        return pa.ipc.RecordBatchReader.from_batches(schema.to_pyarrow(), batches)

    def insert(
        self,
        table_name: str,
        obj: pd.DataFrame | ir.Table | list | dict,
        schema: str | None = None,
        database: str | None = None,
        overwrite: bool = False,
    ) -> None:
        """Insert data into a table.

        Parameters
        ----------
        table_name
            The name of the table to which data needs will be inserted
        obj
            The source data or expression to insert
        schema
            The name of the schema that the table is located in
        database
            Name of the attached database that the table is located in.
        overwrite
            If `True` then replace existing contents of table

        """
        if overwrite:
            self.truncate_table(table_name, schema=schema, database=database)

        if not isinstance(obj, ir.Table):
            obj = ibis.memtable(obj)

        self._run_pre_execute_hooks(obj)

        compiler = self.compiler
        quoted = compiler.quoted
        query = sge.insert(
            expression=self.compile(obj),
            into=sg.table(table_name, db=schema, catalog=database, quoted=quoted),
            columns=[
                sg.to_identifier(col, quoted=quoted)
                for col in self.get_schema(table_name).names
            ],
            dialect=compiler.dialect,
        )

        with self._safe_raw_sql(query):
            pass

    def truncate_table(
        self, name: str, database: str | None = None, schema: str | None = None
    ) -> None:
        """Delete all rows from a table.

        Parameters
        ----------
        name
            Table name
        database
            Database name
        schema
            Schema name

        """
        ident = sg.table(
            name, db=schema, catalog=database, quoted=self.compiler.quoted
        ).sql(self.dialect)
        with self._safe_raw_sql(f"TRUNCATE TABLE {ident}"):
            pass

    def disconnect(self):
        # This is part of the Python DB-API specification so should work for
        # _most_ sqlglot backends
        self.con.close()

    def _compile_builtin_udf(self, udf_node: ops.ScalarUDF | ops.AggUDF) -> None:
        """Compile a built-in UDF. No-op by default."""

    def _compile_python_udf(self, udf_node: ops.ScalarUDF) -> None:
        raise NotImplementedError(
            f"Python UDFs are not supported in the {self.name} backend"
        )

    def _compile_pyarrow_udf(self, udf_node: ops.ScalarUDF) -> None:
        raise NotImplementedError(
            f"PyArrow UDFs are not supported in the {self.name} backend"
        )

    def _compile_pandas_udf(self, udf_node: ops.ScalarUDF) -> str:
        raise NotImplementedError(
            f"pandas UDFs are not supported in the {self.name} backend"
        )



from __future__ import annotations

import contextlib
import inspect
import typing
from collections.abc import Mapping
from pathlib import Path
from typing import TYPE_CHECKING, Any

import datafusion as df
import pyarrow as pa
import pyarrow.dataset as ds
import pyarrow_hotfix  # noqa: F401
import sqlglot as sg
import sqlglot.expressions as sge

import ibis
import ibis.common.exceptions as com
import ibis.expr.datatypes as dt
import ibis.expr.operations as ops
import ibis.expr.schema as sch
import ibis.expr.types as ir
from ibis.backends.base import CanCreateDatabase, CanCreateSchema, NoUrl
from ibis.backends.base.sqlglot import SQLGlotBackend
from ibis.backends.base.sqlglot.compiler import C
from ibis.backends.datafusion.compiler import DataFusionCompiler
from ibis.expr.operations.udf import InputType
from ibis.formats.pyarrow import PyArrowType
from ibis.util import gen_name, normalize_filename

try:
    from datafusion import ExecutionContext as SessionContext
except ImportError:
    from datafusion import SessionContext

try:
    from datafusion import SessionConfig
except ImportError:
    SessionConfig = None

if TYPE_CHECKING:
    from collections.abc import Iterator

    import pandas as pd


class Backend(SQLGlotBackend, CanCreateDatabase, CanCreateSchema, NoUrl):
    name = "datafusion"
    supports_in_memory_tables = True
    supports_arrays = True
    compiler = DataFusionCompiler()

    @property
    def version(self):
        import importlib.metadata

        return importlib.metadata.version("datafusion")

    def do_connect(
        self, config: Mapping[str, str | Path] | SessionContext | None = None
    ) -> None:
        """Create a Datafusion backend for use with Ibis.

        Parameters
        ----------
        config
            Mapping of table names to files.

        Examples
        --------
        >>> import ibis
        >>> config = {"t": "path/to/file.parquet", "s": "path/to/file.csv"}
        >>> ibis.datafusion.connect(config)

        """
        if isinstance(config, SessionContext):
            (self.con, config) = (config, None)
        else:
            if config is not None and not isinstance(config, Mapping):
                raise TypeError("Input to ibis.datafusion.connect must be a mapping")
            if SessionConfig is not None:
                df_config = SessionConfig(
                    {"datafusion.sql_parser.dialect": "PostgreSQL"}
                ).with_information_schema(True)
            else:
                df_config = None
            self.con = SessionContext(df_config)

        self._register_builtin_udfs()

        if not config:
            config = {}

        for name, path in config.items():
            self.register(path, table_name=name)

    def disconnect(self) -> None:
        pass

    @contextlib.contextmanager
    def _safe_raw_sql(self, sql: sge.Statement) -> Any:
        yield self.raw_sql(sql).collect()

    def _metadata(self, query: str) -> Iterator[tuple[str, dt.DataType]]:
        name = gen_name("datafusion_metadata_view")
        table = sg.table(name, quoted=self.compiler.quoted)
        src = sge.Create(
            this=table,
            kind="VIEW",
            expression=sg.parse_one(query, read="datafusion"),
            properties=sge.Properties(expressions=[sge.TemporaryProperty()]),
        )

        with self._safe_raw_sql(src):
            pass

        try:
            result = (
                self.raw_sql(f"DESCRIBE {table.sql(self.name)}")
                .to_arrow_table()
                .to_pydict()
            )
        finally:
            self.drop_view(name)
        return (
            (
                name,
                self.compiler.type_mapper.from_string(
                    type_string, nullable=is_nullable == "YES"
                ),
            )
            for name, type_string, is_nullable in zip(
                result["column_name"], result["data_type"], result["is_nullable"]
            )
        )

    def _register_builtin_udfs(self):
        from ibis.backends.datafusion import udfs

        for name, func in inspect.getmembers(
            udfs,
            predicate=lambda m: callable(m)
            and not m.__name__.startswith("_")
            and m.__module__ == udfs.__name__,
        ):
            annotations = typing.get_type_hints(func)
            argnames = list(inspect.signature(func).parameters.keys())
            input_types = [
                PyArrowType.from_ibis(dt.dtype(annotations.get(arg_name)))
                for arg_name in argnames
            ]
            return_type = PyArrowType.from_ibis(dt.dtype(annotations["return"]))
            udf = df.udf(
                func,
                input_types=input_types,
                return_type=return_type,
                volatility="immutable",
                name=name,
            )
            self.con.register_udf(udf)

    def _register_udfs(self, expr: ir.Expr) -> None:
        for udf_node in expr.op().find(ops.ScalarUDF):
            if udf_node.__input_type__ == InputType.PYARROW:
                udf = self._compile_pyarrow_udf(udf_node)
                self.con.register_udf(udf)

        for udf_node in expr.op().find(ops.ElementWiseVectorizedUDF):
            udf = self._compile_elementwise_udf(udf_node)
            self.con.register_udf(udf)

    def _compile_pyarrow_udf(self, udf_node):
        return df.udf(
            udf_node.__func__,
            input_types=[PyArrowType.from_ibis(arg.dtype) for arg in udf_node.args],
            return_type=PyArrowType.from_ibis(udf_node.dtype),
            volatility=getattr(udf_node, "__config__", {}).get(
                "volatility", "volatile"
            ),
            name=udf_node.__func_name__,
        )

    def _compile_elementwise_udf(self, udf_node):
        return df.udf(
            udf_node.func,
            input_types=list(map(PyArrowType.from_ibis, udf_node.input_type)),
            return_type=PyArrowType.from_ibis(udf_node.return_type),
            volatility="volatile",
            name=udf_node.func.__name__,
        )

    def raw_sql(self, query: str | sge.Expression) -> Any:
        """Execute a SQL string `query` against the database.

        Parameters
        ----------
        query
            Raw SQL string
        kwargs
            Backend specific query arguments

        """
        with contextlib.suppress(AttributeError):
            query = query.sql(dialect=self.dialect, pretty=True)
        self._log(query)
        return self.con.sql(query)

    @property
    def current_database(self) -> str:
        raise NotImplementedError()

    @property
    def current_schema(self) -> str:
        return NotImplementedError()

    def list_databases(self, like: str | None = None) -> list[str]:
        code = (
            sg.select(C.table_catalog)
            .from_(sg.table("tables", db="information_schema"))
            .distinct()
        ).sql()
        result = self.con.sql(code).to_pydict()
        return self._filter_with_like(result["table_catalog"], like)

    def create_database(self, name: str, force: bool = False) -> None:
        with self._safe_raw_sql(
            sge.Create(kind="DATABASE", this=sg.to_identifier(name), exists=force)
        ):
            pass

    def drop_database(self, name: str, force: bool = False) -> None:
        raise com.UnsupportedOperationError(
            "DataFusion does not support dropping databases"
        )

    def list_schemas(
        self, like: str | None = None, database: str | None = None
    ) -> list[str]:
        return self._filter_with_like(
            self.con.catalog(
                database if database is not None else "datafusion"
            ).names(),
            like=like,
        )

    def create_schema(
        self, name: str, database: str | None = None, force: bool = False
    ) -> None:
        # not actually a table, but this is how sqlglot represents schema names
        schema_name = sg.table(name, db=database)
        with self._safe_raw_sql(
            sge.Create(kind="SCHEMA", this=schema_name, exists=force)
        ):
            pass

    def drop_schema(
        self, name: str, database: str | None = None, force: bool = False
    ) -> None:
        schema_name = sg.table(name, db=database)
        with self._safe_raw_sql(
            sge.Drop(kind="SCHEMA", this=schema_name, exists=force)
        ):
            pass

    def list_tables(
        self,
        like: str | None = None,
        database: str | None = None,
    ) -> list[str]:
        """List the available tables."""
        return self._filter_with_like(self.con.tables(), like)

    def get_schema(
        self, table_name: str, schema: str | None = None, database: str | None = None
    ) -> sch.Schema:
        if database is not None:
            catalog = self.con.catalog(database)
        else:
            catalog = self.con.catalog()

        if schema is not None:
            database = catalog.database(schema)
        else:
            database = catalog.database()

        table = database.table(table_name)
        return sch.schema(table.schema)

    def register(
        self,
        source: str | Path | pa.Table | pa.RecordBatch | pa.Dataset | pd.DataFrame,
        table_name: str | None = None,
        **kwargs: Any,
    ) -> ir.Table:
        """Register a data set with `table_name` located at `source`.

        Parameters
        ----------
        source
            The data source(s). May be a path to a file or directory of
            parquet/csv files, a pandas dataframe, or a pyarrow table, dataset
            or record batch.
        table_name
            The name of the table
        kwargs
            Datafusion-specific keyword arguments

        Examples
        --------
        Register a csv:

        >>> import ibis
        >>> conn = ibis.datafusion.connect(config)
        >>> conn.register("path/to/data.csv", "my_table")
        >>> conn.table("my_table")

        Register a PyArrow table:

        >>> import pyarrow as pa
        >>> tab = pa.table({"x": [1, 2, 3]})
        >>> conn.register(tab, "my_table")
        >>> conn.table("my_table")

        Register a PyArrow dataset:

        >>> import pyarrow.dataset as ds
        >>> dataset = ds.dataset("path/to/table")
        >>> conn.register(dataset, "my_table")
        >>> conn.table("my_table")

        """
        import pandas as pd

        if isinstance(source, (str, Path)):
            first = str(source)
        elif isinstance(source, pa.Table):
            self.con.deregister_table(table_name)
            self.con.register_record_batches(table_name, [source.to_batches()])
            return self.table(table_name)
        elif isinstance(source, pa.RecordBatch):
            self.con.deregister_table(table_name)
            self.con.register_record_batches(table_name, [[source]])
            return self.table(table_name)
        elif isinstance(source, pa.dataset.Dataset):
            self.con.deregister_table(table_name)
            self.con.register_dataset(table_name, source)
            return self.table(table_name)
        elif isinstance(source, pd.DataFrame):
            return self.register(pa.Table.from_pandas(source), table_name, **kwargs)
        else:
            raise ValueError("`source` must be either a string or a pathlib.Path")

        if first.startswith(("parquet://", "parq://")) or first.endswith(
            ("parq", "parquet")
        ):
            return self.read_parquet(source, table_name=table_name, **kwargs)
        elif first.startswith(("csv://", "txt://")) or first.endswith(
            ("csv", "tsv", "txt")
        ):
            return self.read_csv(source, table_name=table_name, **kwargs)
        else:
            self._register_failure()
            return None

    def _register_failure(self):
        import inspect

        msg = ", ".join(
            m[0] for m in inspect.getmembers(self) if m[0].startswith("read_")
        )
        raise ValueError(
            f"Cannot infer appropriate read function for input, "
            f"please call one of {msg} directly"
        )

    def _register_in_memory_table(self, op: ops.InMemoryTable) -> None:
        name = op.name
        schema = op.schema

        self.con.deregister_table(name)
        if batches := op.data.to_pyarrow(schema).to_batches():
            self.con.register_record_batches(name, [batches])
        else:
            empty_dataset = ds.dataset([], schema=schema.to_pyarrow())
            self.con.register_dataset(name=name, dataset=empty_dataset)

    def _register_in_memory_tables(self, expr: ir.Expr) -> None:
        if self.supports_in_memory_tables:
            for memtable in expr.op().find(ops.InMemoryTable):
                self._register_in_memory_table(memtable)

    def read_csv(
        self, path: str | Path, table_name: str | None = None, **kwargs: Any
    ) -> ir.Table:
        """Register a CSV file as a table in the current database.

        Parameters
        ----------
        path
            The data source. A string or Path to the CSV file.
        table_name
            An optional name to use for the created table. This defaults to
            a sequentially generated name.
        **kwargs
            Additional keyword arguments passed to Datafusion loading function.

        Returns
        -------
        ir.Table
            The just-registered table

        """
        path = normalize_filename(path)
        table_name = table_name or gen_name("read_csv")
        # Our other backends support overwriting views / tables when reregistering
        self.con.deregister_table(table_name)
        self.con.register_csv(table_name, path, **kwargs)
        return self.table(table_name)

    def read_parquet(
        self, path: str | Path, table_name: str | None = None, **kwargs: Any
    ) -> ir.Table:
        """Register a parquet file as a table in the current database.

        Parameters
        ----------
        path
            The data source.
        table_name
            An optional name to use for the created table. This defaults to
            a sequentially generated name.
        **kwargs
            Additional keyword arguments passed to Datafusion loading function.

        Returns
        -------
        ir.Table
            The just-registered table

        """
        path = normalize_filename(path)
        table_name = table_name or gen_name("read_parquet")
        # Our other backends support overwriting views / tables when reregistering
        self.con.deregister_table(table_name)
        self.con.register_parquet(table_name, path, **kwargs)
        return self.table(table_name)

    def read_delta(
        self, source_table: str | Path, table_name: str | None = None, **kwargs: Any
    ) -> ir.Table:
        """Register a Delta Lake table as a table in the current database.

        Parameters
        ----------
        source_table
            The data source. Must be a directory
            containing a Delta Lake table.
        table_name
            An optional name to use for the created table. This defaults to
            a sequentially generated name.
        **kwargs
            Additional keyword arguments passed to deltalake.DeltaTable.

        Returns
        -------
        ir.Table
            The just-registered table

        """
        source_table = normalize_filename(source_table)

        table_name = table_name or gen_name("read_delta")

        # Our other backends support overwriting views / tables when reregistering
        self.con.deregister_table(table_name)

        try:
            from deltalake import DeltaTable
        except ImportError:
            raise ImportError(
                "The deltalake extra is required to use the "
                "read_delta method. You can install it using pip:\n\n"
                "pip install 'ibis-framework[deltalake]'\n"
            )

        delta_table = DeltaTable(source_table, **kwargs)

        return self.register(delta_table.to_pyarrow_dataset(), table_name=table_name)

    def to_pyarrow_batches(
        self,
        expr: ir.Expr,
        *,
        chunk_size: int = 1_000_000,
        **kwargs: Any,
    ) -> pa.ipc.RecordBatchReader:
        pa = self._import_pyarrow()

        self._register_udfs(expr)
        self._register_in_memory_tables(expr)

        table_expr = expr.as_table()
        raw_sql = self.compile(table_expr, **kwargs)

        frame = self.con.sql(raw_sql)

        schema = table_expr.schema()
        names = schema.names

        struct_schema = schema.as_struct().to_pyarrow()

        return pa.ipc.RecordBatchReader.from_batches(
            schema.to_pyarrow(),
            (
                # convert the renamed + casted columns into a record batch
                pa.RecordBatch.from_struct_array(
                    # rename columns to match schema because datafusion lowercases things
                    pa.RecordBatch.from_arrays(batch.columns, names=names)
                    # cast the struct array to the desired types to work around
                    # https://github.com/apache/arrow-datafusion-python/issues/534
                    .to_struct_array()
                    .cast(struct_schema)
                )
                for batch in frame.collect()
            ),
        )

    def to_pyarrow(self, expr: ir.Expr, **kwargs: Any) -> pa.Table:
        batch_reader = self.to_pyarrow_batches(expr, **kwargs)
        arrow_table = batch_reader.read_all()
        return expr.__pyarrow_result__(arrow_table)

    def execute(self, expr: ir.Expr, **kwargs: Any):
        batch_reader = self.to_pyarrow_batches(expr, **kwargs)
        return expr.__pandas_result__(
            batch_reader.read_pandas(timestamp_as_object=True)
        )

    def create_table(
        self,
        name: str,
        obj: pd.DataFrame | pa.Table | ir.Table | None = None,
        *,
        schema: sch.Schema | None = None,
        database: str | None = None,
        temp: bool = False,
        overwrite: bool = False,
    ):
        """Create a table in Datafusion.

        Parameters
        ----------
        name
            Name of the table to create
        obj
            The data with which to populate the table; optional, but at least
            one of `obj` or `schema` must be specified
        schema
            The schema of the table to create; optional, but at least one of
            `obj` or `schema` must be specified
        database
            The name of the database in which to create the table; if not
            passed, the current database is used.
        temp
            Create a temporary table
        overwrite
            If `True`, replace the table if it already exists, otherwise fail
            if the table exists

        """
        if obj is None and schema is None:
            raise ValueError("Either `obj` or `schema` must be specified")

        properties = []

        if temp:
            properties.append(sge.TemporaryProperty())

        quoted = self.compiler.quoted

        if obj is not None:
            if not isinstance(obj, ir.Expr):
                table = ibis.memtable(obj)
            else:
                table = obj

            self._run_pre_execute_hooks(table)

            relname = "_"
            query = sg.select(
                *(
                    self.compiler.cast(
                        sg.column(col, table=relname, quoted=quoted), dtype
                    ).as_(col, quoted=quoted)
                    for col, dtype in table.schema().items()
                )
            ).from_(
                self._to_sqlglot(table).subquery(
                    sg.to_identifier(relname, quoted=quoted)
                )
            )
        else:
            query = None

        table_ident = sg.to_identifier(name, quoted=quoted)

        if query is None:
            column_defs = [
                sge.ColumnDef(
                    this=sg.to_identifier(colname, quoted=quoted),
                    kind=self.compiler.type_mapper.from_ibis(typ),
                    constraints=(
                        None
                        if typ.nullable
                        else [sge.ColumnConstraint(kind=sge.NotNullColumnConstraint())]
                    ),
                )
                for colname, typ in (schema or table.schema()).items()
            ]

            target = sge.Schema(this=table_ident, expressions=column_defs)
        else:
            target = table_ident

        create_stmt = sge.Create(
            kind="TABLE",
            this=target,
            properties=sge.Properties(expressions=properties),
            expression=query,
            replace=overwrite,
        )

        with self._safe_raw_sql(create_stmt):
            pass

        return self.table(name, schema=database)

    def truncate_table(
        self, name: str, database: str | None = None, schema: str | None = None
    ) -> None:
        """Delete all rows from a table.

        Parameters
        ----------
        name
            Table name
        database
            Database name
        schema
            Schema name

        """
        # datafusion doesn't support `TRUNCATE TABLE` so we use `DELETE FROM`
        #
        # however datafusion as of 34.0.0 doesn't implement DELETE DML yet
        ident = sg.table(name, db=schema, catalog=database).sql(self.name)
        with self._safe_raw_sql(sge.delete(ident)):
            pass



from __future__ import annotations

import ibis
from ibis.expr.types.relations import dereference_mapping

t = ibis.table(
    [
        ("int_col", "int32"),
        ("double_col", "double"),
        ("string_col", "string"),
    ],
    name="t",
)


def dereference_expect(expected):
    return {k.op(): v.op() for k, v in expected.items()}


def test_dereference_project():
    p = t.select([t.int_col, t.double_col])

    mapping = dereference_mapping([p.op()])
    expected = dereference_expect(
        {
            p.int_col: p.int_col,
            p.double_col: p.double_col,
            t.int_col: p.int_col,
            t.double_col: p.double_col,
        }
    )
    assert mapping == expected


def test_dereference_mapping_self_reference():
    v = t.view()

    mapping = dereference_mapping([v.op()])
    expected = dereference_expect(
        {
            v.int_col: v.int_col,
            v.double_col: v.double_col,
            v.string_col: v.string_col,
        }
    )
    assert mapping == expected



from __future__ import annotations

import pyarrow as pa
import pytest
from packaging.version import parse as vparse

pytestmark = pytest.mark.skipif(
    vparse(pa.__version__) < vparse("12"), reason="pyarrow >= 12 required"
)


def test_dataframe_interchange_no_execute(con, alltypes, mocker):
    t = alltypes.select("int_col", "double_col", "string_col")
    pa_df = t.to_pyarrow().__dataframe__()

    to_pyarrow = mocker.spy(con, "to_pyarrow")

    df = t.__dataframe__()

    # Schema metadata
    assert df.num_columns() == pa_df.num_columns()
    assert df.column_names() == pa_df.column_names()

    # Column access
    assert df.get_column(0).dtype == pa_df.get_column(0).dtype
    assert (
        df.get_column_by_name("int_col").dtype
        == pa_df.get_column_by_name("int_col").dtype
    )
    res = [c.dtype for c in df.get_columns()]
    sol = [c.dtype for c in pa_df.get_columns()]
    assert res == sol
    col = df.get_column(0)
    with pytest.raises(
        TypeError, match="only works on a column with categorical dtype"
    ):
        col.describe_categorical  # noqa: B018

    # Subselection
    res = df.select_columns([1, 0])
    sol = pa_df.select_columns([1, 0])
    assert res.column_names() == sol.column_names()
    res = df.select_columns_by_name(["double_col", "int_col"])
    sol = pa_df.select_columns_by_name(["double_col", "int_col"])
    assert res.column_names() == sol.column_names()

    # Nested __dataframe__ access
    df2 = df.__dataframe__()
    pa_df2 = pa_df.__dataframe__()
    assert df2.column_names() == pa_df2.column_names()

    assert not to_pyarrow.called


@pytest.mark.notimpl(["flink"])
def test_dataframe_interchange_dataframe_methods_execute(con, alltypes, mocker):
    t = alltypes.select("int_col", "double_col", "string_col")
    pa_df = t.to_pyarrow().__dataframe__()

    to_pyarrow = mocker.spy(con, "to_pyarrow")

    df = t.__dataframe__()

    assert to_pyarrow.call_count == 0
    assert df.metadata == pa_df.metadata
    assert df.num_rows() == pa_df.num_rows()
    assert df.num_chunks() == pa_df.num_chunks()
    assert len(list(df.get_chunks())) == df.num_chunks()
    assert to_pyarrow.call_count == 1


@pytest.mark.notimpl(["flink"])
def test_dataframe_interchange_column_methods_execute(con, alltypes, mocker):
    t = alltypes.select("int_col", "double_col", "string_col")
    pa_df = t.to_pyarrow().__dataframe__()

    to_pyarrow = mocker.spy(con, "to_pyarrow")

    df = t.__dataframe__()
    col = df.get_column(0)
    pa_col = pa_df.get_column(0)

    assert to_pyarrow.call_count == 0
    assert col.size() == pa_col.size()
    assert col.offset == pa_col.offset

    assert col.describe_null == pa_col.describe_null
    assert col.null_count == pa_col.null_count
    assert col.metadata == pa_col.metadata
    assert col.num_chunks() == pa_col.num_chunks()
    assert len(list(col.get_chunks())) == pa_col.num_chunks()
    assert len(list(col.get_buffers())) == len(list(pa_col.get_buffers()))
    assert to_pyarrow.call_count == 1

    # Access another column doesn't execute
    col2 = df.get_column(1)
    pa_col2 = pa_df.get_column(1)
    assert col2.size() == pa_col2.size()


@pytest.mark.notimpl(["flink"])
def test_dataframe_interchange_select_after_execution_no_reexecute(
    con, alltypes, mocker
):
    t = alltypes.select("int_col", "double_col", "string_col")
    pa_df = t.to_pyarrow().__dataframe__()

    to_pyarrow = mocker.spy(con, "to_pyarrow")

    df = t.__dataframe__()

    # An operation that requires loading data
    assert to_pyarrow.call_count == 0
    assert df.num_rows() == pa_df.num_rows()
    assert to_pyarrow.call_count == 1

    # Subselect columns doesn't reexecute
    df2 = df.select_columns([1, 0])
    pa_df2 = pa_df.select_columns([1, 0])
    assert df2.num_rows() == pa_df2.num_rows()
    assert df2.column_names() == pa_df2.column_names()
    assert to_pyarrow.call_count == 1



from __future__ import annotations

import pytest

import ibis
import ibis.expr.operations as ops
from ibis.common.exceptions import ExpressionError


def test_mutate_with_analytic_functions(alltypes):
    t = alltypes.limit(1000)

    f = t.f
    g = t.group_by(t.g).order_by(t.f)

    exprs = [
        f.lag(),
        f.lead(),
        f.rank(),
        f.dense_rank(),
        f.percent_rank(),
        f.ntile(buckets=7),
        f.first(),
        f.last(),
        f.first().over(ibis.window(preceding=10)),
        f.first().over(ibis.window(following=10)),
        ibis.row_number(),
        f.cumsum(),
        f.cummean(),
        f.cummin(),
        f.cummax(),
        # boolean cumulative reductions
        (f == 0).cumany(),
        (f == 0).cumall(),
        f.sum(),
        f.mean(),
        f.min(),
        f.max(),
    ]

    exprs = [expr.name("e%d" % i) for i, expr in enumerate(exprs)]
    proj = g.mutate(exprs)

    values = list(proj.op().values.values())
    for field in values[len(t.schema()) :]:
        assert isinstance(field, ops.WindowFunction)


def test_value_over_api(alltypes):
    t = alltypes

    w1 = ibis.window(rows=(0, 1), group_by=t.g, order_by=[t.f, t.h])
    w2 = ibis.window(range=(-1, 1), group_by=[t.g, t.a], order_by=[t.f])

    expr = t.f.sum().over(rows=(0, 1), group_by=t.g, order_by=[t.f, t.h])
    expected = t.f.sum().over(w1)
    assert expr.equals(expected)

    expr = t.f.sum().over(range=(-1, 1), group_by=[t.g, t.a], order_by=[t.f])
    expected = t.f.sum().over(w2)
    assert expr.equals(expected)


def test_conflicting_window_boundaries(alltypes):
    t = alltypes

    with pytest.raises(ExpressionError, match="Unable to merge windows"):
        t.f.cumsum().over(rows=(0, 1))


def test_rank_followed_by_over_call_merge_frames(alltypes):
    t = alltypes
    expr1 = t.f.percent_rank().over(ibis.window(group_by=t.f.notnull()))
    expr2 = ibis.percent_rank().over(group_by=t.f.notnull(), order_by=t.f)
    assert expr1.equals(expr2)



from __future__ import annotations

import pytest

import ibis

from .conftest import add_date, tpch_test


@tpch_test
@pytest.mark.notyet(
    ["trino"],
    reason="unreliable due to floating point differences in repeated evaluations of identical subqueries",
    raises=AssertionError,
    strict=False,
)
def test_tpc_h15(lineitem, supplier):
    """Top Supplier Query (Q15)"""

    DATE = "1996-01-01"

    qrev = lineitem
    qrev = qrev.filter(
        [
            lineitem.l_shipdate >= ibis.date(DATE),
            lineitem.l_shipdate < add_date(DATE, dm=3),
        ]
    )

    gqrev = qrev.group_by([lineitem.l_suppkey])
    qrev = gqrev.aggregate(
        total_revenue=(qrev.l_extendedprice * (1 - qrev.l_discount)).sum()
    )

    q = supplier.join(qrev, supplier.s_suppkey == qrev.l_suppkey)
    q = q.filter([q.total_revenue == qrev.total_revenue.max()])
    q = q[q.s_suppkey, q.s_name, q.s_address, q.s_phone, q.total_revenue]
    return q.order_by([q.s_suppkey])



from __future__ import annotations

import uuid

import pytest

import ibis.examples
import ibis.util
from ibis.conftest import CI, LINUX, MACOS, SANDBOXED

pytestmark = pytest.mark.examples

duckdb = pytest.importorskip("duckdb")

# large files or files that are used elsewhere
ignored = frozenset(
    (
        # large
        "imdb_name_basics",
        "imdb_title_akas",
        "imdb_title_basics",
        "imdb_title_crew",
        "imdb_title_episode",
        "imdb_title_principals",
        "imdb_title_ratings",
        "wowah_data_raw",
    )
    * (not CI)  # ignore locally, but not in CI
    + (
        # use in doctests, avoid possible simultaneous use of the downloaded file
        "Aids2",
        "billboard",
        "fish_encounters",
        "penguins",
        "penguins_raw_raw",
        "relig_income_raw",
        "us_rent_income",
        "warpbreaks",
        "who",
        "world_bank_pop_raw",
    )
    * CI  # ignore in CI, but not locally
)

skip_linux_nix = pytest.mark.skipif(
    (LINUX or MACOS) and SANDBOXED,
    reason="nix on linux cannot download duckdb extensions or data due to sandboxing",
)


@pytest.fixture(scope="session")
def example_con(tmp_path_factory, worker_id):
    return ibis.duckdb.connect(
        extension_directory=str(tmp_path_factory.mktemp(worker_id))
    )


@pytest.mark.parametrize("example", sorted(frozenset(dir(ibis.examples)) - ignored))
@pytest.mark.duckdb
@pytest.mark.backend
@skip_linux_nix
def test_examples(example, example_con):
    ex = getattr(ibis.examples, example)

    assert example in repr(ex)

    t = ex.fetch(backend=example_con)
    assert t.op().name == example

    n = t.limit(1).count().execute()
    assert n == 1


def test_non_example():
    gobbledygook = f"{ibis.util.guid()}"
    with pytest.raises(AttributeError, match=gobbledygook):
        getattr(ibis.examples, gobbledygook)


@pytest.mark.duckdb
@pytest.mark.backend
@skip_linux_nix
def test_backend_arg():
    con = ibis.duckdb.connect()
    t = ibis.examples.penguins.fetch(backend=con)
    assert t.get_name() in con.list_tables()


@pytest.mark.duckdb
@pytest.mark.backend
@skip_linux_nix
def test_table_name_arg():
    con = ibis.duckdb.connect()
    name = f"penguins-{uuid.uuid4().hex}"
    t = ibis.examples.penguins.fetch(backend=con, table_name=name)
    assert t.get_name() == name



from __future__ import annotations

from typing import TYPE_CHECKING

import sqlglot as sg

import ibis.common.exceptions as exc
import ibis.expr.schema as sch
from ibis.backends.base.sqlglot.datatypes import FlinkType
from ibis.backends.base.sqlglot.ddl import DDL, DML, CreateDDL, DropObject
from ibis.util import promote_list

if TYPE_CHECKING:
    from collections.abc import Sequence

    from ibis.expr.api import Watermark


class FlinkBase:
    dialect = "hive"

    def format_dtype(self, dtype):
        sql_string = FlinkType.from_ibis(dtype)
        if dtype.is_timestamp():
            return (
                f"TIMESTAMP({dtype.scale})" if dtype.scale is not None else "TIMESTAMP"
            )
        else:
            return sql_string.sql("flink") + " NOT NULL" * (not dtype.nullable)

    def format_properties(self, props):
        tokens = []
        for k, v in sorted(props.items()):
            tokens.append(f"  '{k}'='{v}'")
        return "(\n{}\n)".format(",\n".join(tokens))

    def format_watermark_strategy(self, watermark: Watermark) -> str:
        from ibis.backends.flink.utils import translate_literal

        if watermark.allowed_delay is None:
            return watermark.time_col
        return (
            f"{watermark.time_col} - {translate_literal(watermark.allowed_delay.op())}"
        )

    def format_schema_with_watermark(
        self,
        schema: sch.Schema,
        watermark: Watermark | None = None,
        primary_keys: Sequence[str] | None = None,
    ) -> str:
        elements = [
            f"{self.quote(name)} {self.format_dtype(t)}"
            for name, t in zip(schema.names, schema.types)
        ]

        if watermark is not None:
            elements.append(
                f"WATERMARK FOR {watermark.time_col} AS {self.format_watermark_strategy(watermark)}"
            )

        if primary_keys is not None and primary_keys:
            # Note (mehmet): Currently supports "NOT ENFORCED" only. For the reason
            # of this choice, the following quote from Flink docs is self-explanatory:
            # "SQL standard specifies that a constraint can either be ENFORCED or
            # NOT ENFORCED. This controls if the constraint checks are performed on
            # the incoming/outgoing data. Flink does not own the data therefore the
            # only mode we want to support is the NOT ENFORCED mode. It is up to the
            # user to ensure that the query enforces key integrity."
            # Ref: https://nightlies.apache.org/flink/flink-docs-release-1.18/docs/dev/table/sql/create/#primary-key
            comma_separated_keys = ", ".join(f"`{key}`" for key in primary_keys)
            elements.append(f"PRIMARY KEY ({comma_separated_keys}) NOT ENFORCED")

        return "({})".format(",\n ".join(elements))


class CreateTableWithSchema(FlinkBase, CreateDDL):
    def __init__(
        self,
        table_name: str,
        schema: sch.Schema,
        database=None,
        catalog=None,
        can_exist=False,
        external=False,
        partition=None,
        primary_key: str | Sequence[str] | None = None,
        tbl_properties=None,
        temporary: bool = False,
        watermark: Watermark | None = None,
    ):
        self.can_exist = can_exist
        self.catalog = catalog
        self.database = database
        self.partition = partition
        self.primary_keys = promote_list(primary_key)
        self.schema = schema
        self.table_name = table_name
        self.tbl_properties = tbl_properties
        self.temporary = temporary
        self.watermark = watermark

        # Check if `primary_keys` is a subset of the columns in `schema`.
        if self.primary_keys and not set(self.primary_keys) <= set(schema.names):
            raise exc.IbisError(
                "`primary_key` must be a subset of the columns in `schema`. \n"
                f"\t primary_key= {primary_key} \n"
                f"\t schema.names= {schema.names}"
            )

    @property
    def _prefix(self) -> str:
        # `TEMPORARY` is not documented in Flink's documentation
        modifier = " TEMPORARY" if self.temporary else ""
        return f"CREATE{modifier} TABLE"

    def _create_line(self) -> str:
        scoped_name = self.scoped_name(self.table_name, self.database, self.catalog)
        return f"{self._prefix} {self._if_exists()}{scoped_name}"

    @property
    def _pieces(self):
        if self.partition is not None:
            main_schema = self.schema
            part_schema = self.partition
            if not isinstance(part_schema, sch.Schema):
                part_fields = {name: self.schema[name] for name in part_schema}
                part_schema = sch.Schema(part_fields)

            to_delete = {name for name in self.partition if name in self.schema}
            fields = {
                name: dtype
                for name, dtype in main_schema.items()
                if name not in to_delete
            }
            main_schema = sch.Schema(fields)

            yield self.format_schema_with_watermark(
                main_schema, self.watermark, self.primary_keys
            )
            yield f"PARTITIONED BY {self.format_schema(part_schema)}"
        else:
            yield self.format_schema_with_watermark(
                self.schema, self.watermark, self.primary_keys
            )

        yield f"WITH {self.format_properties(self.tbl_properties)}"

    @property
    def pieces(self):
        yield self._create_line()
        yield from filter(None, self._pieces)

    def compile(self):
        return "\n".join(self.pieces)


class CreateView(FlinkBase, CreateDDL):
    def __init__(
        self,
        name: str,
        query_expression: str,
        database: str | None = None,
        catalog: str | None = None,
        can_exist: bool = False,
        temporary: bool = False,
    ):
        super().__init__(
            table_name=name,
            database=database,
            can_exist=can_exist,
        )
        self.name = name
        self.query_expression = query_expression
        self.catalog = catalog
        self.temporary = temporary

    @property
    def _prefix(self):
        if self.temporary:
            return "CREATE TEMPORARY VIEW"
        else:
            return "CREATE VIEW"

    def _create_line(self):
        scoped_name = self.scoped_name(self.name, self.database, self.catalog)
        return f"{self._prefix} {self._if_exists()}{scoped_name}"

    @property
    def pieces(self):
        yield self._create_line()
        yield f"AS {self.query_expression}"

    def compile(self):
        return "\n".join(self.pieces)


class DropTable(FlinkBase, DropObject):
    _object_type = "TABLE"

    def __init__(
        self,
        table_name: str,
        database: str | None = None,
        catalog: str | None = None,
        must_exist: bool = True,
        temporary: bool = False,
    ):
        super().__init__(must_exist=must_exist)
        self.table_name = table_name
        self.database = database
        self.catalog = catalog
        self.temporary = temporary

    def _object_name(self):
        return self.scoped_name(self.table_name, self.database, self.catalog)

    def compile(self):
        temporary = "TEMPORARY " if self.temporary else ""
        if_exists = "" if self.must_exist else "IF EXISTS "
        object_name = self._object_name()
        return f"DROP {temporary}{self._object_type} {if_exists}{object_name}"


class DropView(DropTable):
    _object_type = "VIEW"

    def __init__(
        self,
        name: str,
        database: str | None = None,
        catalog: str | None = None,
        must_exist: bool = True,
        temporary: bool = False,
    ):
        super().__init__(
            table_name=name,
            database=database,
            catalog=catalog,
            must_exist=must_exist,
            temporary=temporary,
        )


class RenameTable(FlinkBase, DDL):
    def __init__(self, old_name: str, new_name: str, must_exist: bool = True):
        self.old_name = old_name
        self.new_name = new_name
        self.must_exist = must_exist

    def compile(self):
        if_exists = "" if self.must_exist else "IF EXISTS"
        return f"ALTER TABLE {if_exists} {self.old_name} RENAME TO {self.new_name}"


class _DatabaseObject:
    def _object_name(self):
        name = sg.to_identifier(self.name, quoted=True).sql(dialect=self.dialect)
        if self.catalog:
            catalog = sg.to_identifier(self.catalog, quoted=True).sql(
                dialect=self.dialect
            )
            return f"{catalog}.{name}"
        else:
            return name


class CreateDatabase(FlinkBase, _DatabaseObject, CreateDDL):
    def __init__(
        self,
        name: str,
        db_properties: dict | None,
        catalog: str | None = None,
        can_exist: bool = False,
    ):
        # TODO(chloeh13q): support COMMENT
        self.name = name
        self.db_properties = db_properties
        self.catalog = catalog
        self.can_exist = can_exist

    def _format_db_properties(self) -> str:
        return (
            f"WITH {self.format_properties(self.db_properties)}"
            if self.db_properties
            else ""
        )

    def compile(self):
        create_decl = "CREATE DATABASE"
        create_line = f"{create_decl} {self._if_exists()}{self._object_name()}"

        return f"{create_line}\n{self._format_db_properties()}"


class DropDatabase(FlinkBase, _DatabaseObject, DropObject):
    _object_type = "DATABASE"

    def __init__(self, name: str, catalog: str | None = None, must_exist: bool = True):
        super().__init__(must_exist=must_exist)
        self.name = name
        self.catalog = catalog


class InsertSelect(FlinkBase, DML):
    def __init__(
        self,
        table_name,
        select_expr,
        database: str | None = None,
        catalog: str | None = None,
        partition=None,
        partition_schema=None,
        overwrite=False,
    ):
        self.table_name = table_name
        self.database = database
        self.catalog = catalog
        self.select = select_expr
        self.partition = partition
        self.partition_schema = partition_schema
        self.overwrite = overwrite

    def compile(self):
        if self.overwrite:
            cmd = "INSERT OVERWRITE"
        else:
            cmd = "INSERT INTO"

        if self.partition is not None:
            part = self.format_partition(self.partition, self.partition_schema)
            partition = f" {part} "
        else:
            partition = ""

        select_query = self.select
        scoped_name = self.scoped_name(self.table_name, self.database, self.catalog)
        return f"{cmd} {scoped_name}{partition}\n{select_query}"



from __future__ import annotations

from decimal import Context, localcontext
from decimal import Decimal as D

import pytest

from ibis.common.numeric import normalize_decimal


@pytest.mark.parametrize(
    ("value", "precision", "scale", "expected"),
    [
        (1, None, None, D("1")),
        (1.0, None, None, D("1.0")),
        (1.0, 2, None, D("1.0")),
        (1.0, 2, 1, D("1.0")),
        (1.0, 3, 2, D("1.0")),
        (1.0, 3, 1, D("1.0")),
        (1.0, 3, 0, D("1")),
        (1.0, 2, 0, D("1")),
        (1.0, 1, 0, D("1")),
        (3.14, 3, 2, D("3.14")),
        (3.14, 10, 2, D("3.14")),
        (3.14, 10, 3, D("3.14")),
        (3.14, 10, 4, D("3.14")),
        (1234.567, 10, 4, D("1234.567")),
        (1234.567, 10, 3, D("1234.567")),
    ],
)
def test_normalize_decimal(value, precision, scale, expected):
    assert normalize_decimal(value, precision, scale) == expected


@pytest.mark.parametrize(
    ("value", "precision", "scale"),
    [
        (1.0, 2, 2),
        (1.0, 1, 1),
        (D("1.1234"), 5, 3),
        (D("1.1234"), 4, 2),
        (D("23145"), 4, 2),
        (1234.567, 10, 2),
        (1234.567, 10, 1),
        (3.14, 10, 0),
        (3.14, 3, 0),
        (3.14, 3, 1),
        (3.14, 10, 1),
    ],
)
def test_normalize_failing(value, precision, scale):
    with pytest.raises(TypeError):
        normalize_decimal(value, precision, scale)


def test_normalize_decimal_dont_truncate_precision():
    # test that the decimal context is ignored, 38 is the default precision
    for prec in [10, 30, 38]:
        with localcontext(Context(prec=prec)):
            v = "1.123456789"
            assert str(normalize_decimal(v + "0000")) == "1.123456789"

            v = v + "1" * 28
            assert len(v) == 39
            assert str(normalize_decimal(v)) == v

            # if no precision is specified, we use precision 38 for dec.normalize()
            v = v + "1"
            assert len(v) == 40
            assert str(normalize_decimal(v)) == v[:-1]

            # pass the precision explicitly
            assert str(normalize_decimal(v, precision=39)) == v

            v = v + "1" * 11
            assert len(v) == 51
            assert str(normalize_decimal(v, precision=50)) == v
            assert str(normalize_decimal(v, precision=45)) == v[:-5]



from __future__ import annotations

import hypothesis as h
import hypothesis.strategies as st
import pyarrow as pa
import pyarrow.tests.strategies as past
import pytest

import ibis.expr.datatypes as dt
from ibis.common.exceptions import IntegrityError
from ibis.formats.pyarrow import PyArrowSchema, PyArrowType


def assert_dtype_roundtrip(arrow_type, ibis_type=None, restored_type=None):
    dtype = PyArrowType.to_ibis(arrow_type, nullable=False)
    if ibis_type is not None:
        assert dtype == ibis_type

    patyp = PyArrowType.from_ibis(dtype)
    if restored_type is None:
        restored_type = arrow_type
    assert patyp == restored_type


roundtripable_types = st.deferred(
    lambda: (
        past.null_type
        | past.bool_type
        | past.integer_types
        | past.floating_types
        | past.duration_types
        | past.string_type
        | past.binary_type
        | past.timestamp_types
        | st.builds(pa.list_, roundtripable_types)
        | past.struct_types(roundtripable_types)
        | past.map_types(roundtripable_types, roundtripable_types)
    )
)


@h.given(roundtripable_types)
def test_roundtripable_types(arrow_type):
    assert_dtype_roundtrip(arrow_type)


@pytest.mark.parametrize(
    ("arrow_type", "ibis_type", "restored_type"),
    [
        (pa.decimal128(1, 1), dt.Decimal(1, 1, nullable=False), pa.decimal128(1, 1)),
        (pa.decimal128(10, 3), dt.Decimal(10, 3, nullable=False), pa.decimal128(10, 3)),
        (pa.decimal128(38, 3), dt.Decimal(38, 3, nullable=False), pa.decimal128(38, 3)),
        (pa.decimal256(1, 1), dt.Decimal(1, 1, nullable=False), pa.decimal128(1, 1)),
        (pa.decimal256(38, 5), dt.Decimal(38, 5, nullable=False), pa.decimal128(38, 5)),
        (pa.decimal256(39, 6), dt.Decimal(39, 6, nullable=False), pa.decimal256(39, 6)),
        (pa.decimal256(76, 6), dt.Decimal(76, 6, nullable=False), pa.decimal256(76, 6)),
        (pa.date32(), dt.Date(nullable=False), pa.date32()),
        (pa.date64(), dt.Date(nullable=False), pa.date32()),
        (pa.time32("s"), dt.Time(nullable=False), pa.time64("ns")),
        (pa.time32("ms"), dt.Time(nullable=False), pa.time64("ns")),
        (pa.time64("us"), dt.Time(nullable=False), pa.time64("ns")),
        (pa.time64("ns"), dt.Time(nullable=False), pa.time64("ns")),
        (pa.large_binary(), dt.Binary(nullable=False), pa.binary()),
        (pa.large_string(), dt.String(nullable=False), pa.string()),
        (
            pa.large_list(pa.int64()),
            dt.Array(dt.Int64(nullable=True), nullable=False),
            pa.list_(pa.int64()),
        ),
        (
            pa.list_(pa.int64(), list_size=3),
            dt.Array(dt.Int64(nullable=True), nullable=False),
            pa.list_(pa.int64()),
        ),
    ],
)
def test_non_roundtripable_types(arrow_type, ibis_type, restored_type):
    assert_dtype_roundtrip(arrow_type, ibis_type, restored_type)


@pytest.mark.parametrize("timezone", [None, "UTC"])
@pytest.mark.parametrize("nullable", [True, False])
def test_timestamp_no_scale(timezone, nullable):
    dtype = dt.Timestamp(scale=None, timezone=timezone, nullable=nullable)
    assert dtype.to_pyarrow() == pa.timestamp("us", tz=timezone)


def test_month_day_nano_type_unsupported():
    with pytest.raises(ValueError, match="Arrow interval type is not supported"):
        PyArrowType.to_ibis(pa.month_day_nano_interval())


@pytest.mark.parametrize("value_nullable", [True, False])
def test_dtype_from_nullable_map_type(value_nullable):
    # the key field cannot be nullable
    pyarrow_type = pa.map_(
        pa.int64(), pa.field("value", pa.int64(), nullable=value_nullable)
    )
    ibis_type = PyArrowType.to_ibis(pyarrow_type)
    restored_type = PyArrowType.from_ibis(ibis_type)

    assert ibis_type == dt.Map(
        dt.Int64(nullable=False), dt.Int64(nullable=value_nullable)
    )
    assert restored_type.key_field.type == pa.int64()
    assert restored_type.key_field.nullable is False
    assert restored_type.item_field.type == pa.int64()
    assert restored_type.item_field.nullable is value_nullable


@pytest.mark.parametrize("value_nullable", [True, False])
@pytest.mark.parametrize("list_nullable", [True, False])
def test_dtype_from_nullable_list_type(value_nullable, list_nullable):
    pyarrow_type = pa.list_(pa.field("value", pa.int64(), nullable=value_nullable))
    ibis_type = PyArrowType.to_ibis(pyarrow_type, nullable=list_nullable)
    restored_type = PyArrowType.from_ibis(ibis_type)

    assert ibis_type == dt.Array(
        dt.Int64(nullable=value_nullable), nullable=list_nullable
    )
    assert restored_type.value_field.type == pa.int64()
    assert restored_type.value_field.nullable is value_nullable


@pytest.mark.parametrize("value_type", [pa.string(), pa.date32()])
def test_dtype_from_dictionary_type(value_type):
    dict_type = pa.dictionary(pa.int32(), value_type)
    ibis_type = PyArrowType.to_ibis(dict_type)
    expected = PyArrowType.to_ibis(value_type)
    assert ibis_type == expected


@pytest.mark.parametrize(
    ("ibis_type", "arrow_type"),
    [
        (dt.Set(dt.String(nullable=True)), pa.list_(pa.string())),
        (
            dt.Set(dt.String(nullable=False)),
            pa.list_(pa.field("item", pa.string(), nullable=False)),
        ),
        (dt.uuid, pa.string()),
    ],
)
def test_ibis_exclusive_types(ibis_type, arrow_type):
    assert PyArrowType.from_ibis(ibis_type) == arrow_type


def test_schema_from_pyarrow_checks_duplicate_column_names():
    arrow_schema = pa.schema(
        [
            pa.field("a", pa.int64()),
            pa.field("a", pa.int64()),
        ]
    )
    with pytest.raises(IntegrityError, match="Duplicate column name"):
        PyArrowSchema.to_ibis(arrow_schema)


@h.given(past.schemas(roundtripable_types))
def test_schema_roundtrip(pyarrow_schema):
    unique_column_names = set(pyarrow_schema.names)
    h.assume(len(unique_column_names) == len(pyarrow_schema.names))

    ibis_schema = PyArrowSchema.to_ibis(pyarrow_schema)
    restored = PyArrowSchema.from_ibis(ibis_schema)
    assert pyarrow_schema.equals(restored)


def test_unknown_dtype_gets_converted_to_string():
    assert PyArrowType.from_ibis(dt.unknown) == pa.string()


@pytest.mark.parametrize(
    "ibis_type",
    [
        pytest.param(dt.geometry, id="geometry"),
        pytest.param(dt.geography, id="geography"),
        pytest.param(dt.point, id="point"),
        pytest.param(dt.linestring, id="linestring"),
        pytest.param(dt.polygon, id="polygon"),
        pytest.param(dt.multilinestring, id="multilinestring"),
        pytest.param(dt.multipoint, id="multipoint"),
        pytest.param(dt.multipolygon, id="multipolygon"),
    ],
)
def test_geo_gets_converted_to_binary(ibis_type):
    assert PyArrowType.from_ibis(ibis_type) == pa.binary()



from ibis.common.grounds import Annotable
from ibis.common.patterns import InstanceOf

IsAny = InstanceOf(object)
IsBool = InstanceOf(bool)
IsFloat = InstanceOf(float)
IsInt = InstanceOf(int)
IsStr = InstanceOf(str)


class Node(Annotable):
    pass


class Literal(Node):
    value = InstanceOf((int, float, bool, str))
    dtype = InstanceOf(type)

    def __add__(self, other):
        return Add(self, other)


class BinaryOperation(Annotable):
    left = InstanceOf(Node)
    right = InstanceOf(Node)


class Add(BinaryOperation):
    pass


one = Literal(value=1, dtype=int)
two = Literal(value=2, dtype=int)


def test_pattern_matching():
    match one:
        case Literal(value, dtype=dtype):
            assert value == 1
            assert dtype is int
        case _:
            raise ValueError("Unable to match")

    match (one + two):
        case Add(left, right):
            assert left == one
            assert right == two
        case _:
            raise ValueError("Unable to match")



from __future__ import annotations

import os

import pytest

import ibis
import ibis.expr.datashape as ds
import ibis.expr.datatypes as dt
import ibis.expr.operations as ops
import ibis.expr.types as ir

graphviz = pytest.importorskip("graphviz")

import ibis.expr.visualize as viz  # noqa: E402
from ibis.expr import api  # noqa: E402

pytestmark = pytest.mark.skipif(
    int(os.environ.get("CONDA_BUILD", 0)) == 1, reason="CONDA_BUILD defined"
)


def key(node):
    return str(hash(node))


@pytest.mark.parametrize(
    "expr_func",
    [
        lambda t: t.a,
        lambda t: t.a + t.b,
        lambda t: t.a + t.b > 3**t.a,
        lambda t: t[(t.a + t.b * 2 * t.b / t.b**3 > 4) & (t.b > 5)],
        lambda t: t[(t.a + t.b * 2 * t.b / t.b**3 > 4) & (t.b > 5)]
        .group_by("c")
        .aggregate(amean=lambda f: f.a.mean(), bsum=lambda f: f.b.sum()),
    ],
)
def test_exprs(alltypes, expr_func):
    expr = expr_func(alltypes)
    graph = viz.to_graph(expr)
    assert key(alltypes.op()) in graph.source
    assert key(expr.op()) in graph.source


def test_custom_expr():
    class MyExpr(ir.Expr):
        pass

    class MyExprNode(ops.Node):
        foo: ops.Value[dt.String, ds.Any]
        bar: ops.Value[dt.Numeric, ds.Any]

        def to_expr(self):
            return MyExpr(self)

    op = MyExprNode("Hello!", 42.3)
    expr = op.to_expr()
    graph = viz.to_graph(expr)
    assert key(expr.op()) in graph.source


def test_custom_expr_with_not_implemented_type():
    class MyExpr(ir.Expr):
        def type(self):
            raise NotImplementedError

        def schema(self):
            raise NotImplementedError

    class MyExprNode(ops.Node):
        foo: ops.Value[dt.String, ds.Any]
        bar: ops.Value[dt.Numeric, ds.Any]

        def to_expr(self):
            return MyExpr(self)

    op = MyExprNode("Hello!", 42.3)
    expr = op.to_expr()
    graph = viz.to_graph(expr)
    assert key(expr.op()) in graph.source


@pytest.mark.parametrize("how", ["inner", "left", "right", "outer"])
def test_join(how):
    left = ibis.table([("a", "int64"), ("b", "string")])
    right = ibis.table([("b", "string"), ("c", "int64")])
    joined = left.join(right, left.b == right.b, how=how)
    result = joined[left.a, right.c]
    graph = viz.to_graph(result)
    assert key(result.op()) in graph.source


def test_order_by():
    t = ibis.table([("a", "int64"), ("b", "string"), ("c", "int32")])
    expr = t.group_by(t.b).aggregate(sum_a=t.a.sum().cast("double")).order_by("b")
    graph = viz.to_graph(expr)
    assert key(expr.op()) in graph.source


def test_optional_graphviz_repr(monkeypatch):
    monkeypatch.setattr(ibis.options, "graphviz_repr", True)

    t = ibis.table([("a", "int64"), ("b", "string"), ("c", "int32")])
    expr = t.group_by(t.b).aggregate(sum_a=t.a.sum().cast("double")).order_by("b")

    # default behavior
    assert expr._repr_png_() is not None

    # turn it off
    ibis.options.graphviz_repr = False
    assert expr._repr_png_() is None

    # turn it back on
    ibis.options.graphviz_repr = True
    assert expr._repr_png_() is not None


def test_between():
    t = ibis.table([("a", "int64"), ("b", "string"), ("c", "int32")])
    expr = t.a.between(1, 1)
    lower_bound, upper_bound = expr.op().args[1:]
    graph = viz.to_graph(expr)
    source = graph.source

    # one for the node itself and one for the edge to between
    assert key(lower_bound) in source
    assert key(upper_bound) in source


def test_asof_join():
    left = ibis.table([("time", "int32"), ("value", "double")])
    right = ibis.table([("time", "int32"), ("value2", "double")])
    right = right.mutate(foo=1)

    joined = api.asof_join(left, right, "time")
    result = joined[left, right.foo]
    graph = viz.to_graph(result)
    assert key(result.op()) in graph.source


def test_filter():
    # Smoketest that NodeList nodes are handled properly
    t = ibis.table([("a", "int64"), ("b", "string"), ("c", "int32")])
    expr = t.filter((t.a == 1) & (t.b == "x"))
    graph = viz.to_graph(expr, label_edges=True)
    assert "predicates[0]" in graph.source
    assert "predicates[1]" in graph.source


def test_html_escape(monkeypatch):
    monkeypatch.setattr(ibis.options, "graphviz_repr", True)
    # Check that we correctly escape HTML <> characters in the graphviz
    # representation. If an error is thrown, _repr_png_ returns None.
    expr = ibis.table([("<a & b>", ibis.expr.datatypes.Array("string"))])
    assert expr._repr_png_() is not None

    expr = ibis.array([1, 2, 3])
    assert expr._repr_png_() is not None

    expr = ibis.array([1, 2, 3]).name("COL")
    assert expr._repr_png_() is not None



from __future__ import annotations

import pytest

from ibis.common.annotations import attribute
from ibis.common.collections import frozendict
from ibis.common.grounds import Annotable, Concrete

pytestmark = pytest.mark.benchmark


class MyObject(Concrete):
    a: int
    b: str
    c: tuple[int, ...]
    d: frozendict[str, int]

    @attribute
    def e(self):
        return self.a * 2

    @attribute
    def f(self):
        return self.b * 2

    @attribute
    def g(self):
        return self.c * 2


def test_concrete_construction(benchmark):
    benchmark(MyObject, 1, "2", c=(3, 4), d=frozendict(e=5, f=6))


def test_concrete_isinstance(benchmark):
    def check(obj):
        for _ in range(100):
            assert isinstance(obj, MyObject)
            assert isinstance(obj, Concrete)
            assert isinstance(obj, Annotable)

    obj = MyObject(1, "2", c=(3, 4), d=frozendict(e=5, f=6))
    benchmark(check, obj)



from __future__ import annotations

from typing import Any

import pytest

import ibis
from ibis.backends.conftest import TEST_TABLES
from ibis.backends.tests.base import BackendTest
from ibis.backends.tests.data import array_types, win


class TestConf(BackendTest):
    # check_names = False
    # supports_divide_by_zero = True
    # returned_timestamp_unit = 'ns'
    supports_structs = False
    supports_json = False
    supports_arrays = True
    stateful = False
    deps = ("datafusion",)

    def _load_data(self, **_: Any) -> None:
        con = self.connection
        for table_name in TEST_TABLES:
            path = self.data_dir / "parquet" / f"{table_name}.parquet"
            con.register(path, table_name=table_name)
        con.register(array_types, table_name="array_types")
        con.register(win, table_name="win")

    @staticmethod
    def connect(*, tmpdir, worker_id, **kw):
        return ibis.datafusion.connect(**kw)


@pytest.fixture(scope="session")
def con(data_dir, tmp_path_factory, worker_id):
    return TestConf.load_data(data_dir, tmp_path_factory, worker_id).connection


@pytest.fixture(scope="session")
def alltypes(con):
    return con.table("functional_alltypes")


@pytest.fixture(scope="session")
def alltypes_df(alltypes):
    return alltypes.execute()



from __future__ import annotations

import math
from datetime import date, datetime
from operator import methodcaller

import pandas as pd
import pandas.testing as tm
import pytest
from pytest import param

import ibis
import ibis.expr.datatypes as dt
import ibis.expr.types as ir
from ibis import literal as L
from ibis import udf

pytest.importorskip("clickhouse_connect")


@pytest.mark.parametrize("to_type", ["int8", "int16", "float32", "float", "!float64"])
def test_cast_double_col(alltypes, to_type, snapshot):
    expr = alltypes.double_col.cast(to_type)
    result = expr.compile()
    snapshot.assert_match(result, "out.sql")


@pytest.mark.parametrize(
    "to_type",
    [
        "int8",
        "int16",
        "!string",
        "timestamp",
        "date",
        "!map<string, int64>",
        "!struct<a: string, b: int64>",
    ],
)
def test_cast_string_col(alltypes, to_type, snapshot):
    expr = alltypes.string_col.cast(to_type)
    snapshot.assert_match(expr.compile(), "out.sql")


@pytest.mark.parametrize(
    "column",
    [
        "id",
        "bool_col",
        "tinyint_col",
        "smallint_col",
        "int_col",
        "bigint_col",
        "float_col",
        "double_col",
        "date_string_col",
        "string_col",
        "timestamp_col",
        "year",
        "month",
    ],
)
def test_noop_cast(alltypes, column, snapshot):
    col = alltypes[column]
    result = col.cast(col.type())
    assert result.equals(col)
    snapshot.assert_match(result.compile(), "out.sql")


def test_timestamp_cast(alltypes, snapshot):
    target = dt.Timestamp(nullable=False)
    result1 = alltypes.timestamp_col.cast(target)
    result2 = alltypes.int_col.cast(target)

    assert isinstance(result1, ir.TimestampColumn)
    assert isinstance(result2, ir.TimestampColumn)

    snapshot.assert_match(result1.compile(), "out1.sql")
    snapshot.assert_match(result2.compile(), "out2.sql")


def test_timestamp_now(con, snapshot):
    expr = ibis.now()
    snapshot.assert_match(con.compile(expr), "out.sql")


@pytest.mark.parametrize("unit", ["y", "m", "d", "w", "h", "minute"])
def test_timestamp_truncate(con, unit, snapshot):
    stamp = ibis.timestamp("2009-05-17 12:34:56")
    expr = stamp.truncate(unit)
    snapshot.assert_match(con.compile(expr), "out.sql")


@pytest.mark.parametrize(("value", "expected"), [(0, None), (5.5, 5.5)])
def test_nullif_zero(con, value, expected):
    result = con.execute(L(value).nullif(0))
    if expected is None:
        assert pd.isnull(result)
    else:
        assert result == expected


@pytest.mark.parametrize(
    ("expr", "expected"),
    [
        (L(None).isnull(), True),
        (L(1).isnull(), False),
        (L(None).notnull(), False),
        (L(1).notnull(), True),
    ],
)
def test_isnull_notnull(con, expr, expected):
    assert con.execute(expr) == expected


@pytest.mark.parametrize(
    ("expr", "expected"),
    [
        (ibis.coalesce(5, None, 4), 5),
        (ibis.coalesce(ibis.NA, 4, ibis.NA), 4),
        (ibis.coalesce(ibis.NA, ibis.NA, 3.14), 3.14),
    ],
)
def test_coalesce(con, expr, expected):
    assert con.execute(expr) == expected


@pytest.mark.parametrize(
    ("expr", "expected"),
    [
        (ibis.NA.fillna(5), 5),
        (L(5).fillna(10), 5),
        (L(5).nullif(5), None),
        (L(10).nullif(5), 10),
    ],
)
def test_fillna_nullif(con, expr, expected):
    result = con.execute(expr)
    if expected is None:
        assert pd.isnull(result)
    else:
        assert result == expected


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        (L("foo_bar"), "String"),
        (L(5), "UInt8"),
        (L(1.2345), "Float64"),
        (L(datetime(2015, 9, 1, hour=14, minute=48, second=5)), "DateTime"),
        (L(date(2015, 9, 1)), "Date"),
        param(
            ibis.NA,
            "Null",
            marks=pytest.mark.xfail(
                raises=AssertionError,
                reason=(
                    "Client/server version mismatch not handled in the "
                    "clickhouse driver"
                ),
            ),
        ),
    ],
)
def test_typeof(con, value, expected):
    assert con.execute(value.typeof()) == expected


@pytest.mark.parametrize(("value", "expected"), [("foo_bar", 7), ("", 0)])
def test_tuple_string_length(con, value, expected):
    assert con.execute(L(value).length()) == expected


@pytest.mark.parametrize(
    ("op", "expected"),
    [
        (methodcaller("substr", 0, 3), "foo"),
        (methodcaller("substr", 4, 3), "bar"),
        (methodcaller("substr", 1), "oo_bar"),
    ],
)
def test_string_substring(con, op, expected):
    value = L("foo_bar")
    assert con.execute(op(value)) == expected


def test_string_column_substring(con, alltypes, snapshot):
    expr = alltypes.string_col.substr(2)
    snapshot.assert_match(expr.compile(), "out1.sql")
    assert len(con.execute(expr))

    expr = alltypes.string_col.substr(0, 3)
    snapshot.assert_match(expr.compile(), "out2.sql")
    assert len(con.execute(expr))


def test_string_reverse(con):
    assert con.execute(L("foo").reverse()) == "oof"


def test_string_upper(con):
    assert con.execute(L("foo").upper()) == "FOO"


def test_string_lower(con):
    assert con.execute(L("FOO").lower()) == "foo"


def test_string_length(con):
    assert con.execute(L("FOO").length()) == 3


@pytest.mark.parametrize(
    ("value", "op", "expected"),
    [
        (L("foobar"), methodcaller("contains", "bar"), True),
        (L("foobar"), methodcaller("contains", "foo"), True),
        (L("foobar"), methodcaller("contains", "baz"), False),
        (L("100%"), methodcaller("contains", "%"), True),
        (L("a_b_c"), methodcaller("contains", "_"), True),
    ],
)
def test_string_contains(con, op, value, expected):
    assert con.execute(op(value)) == expected


def test_re_replace(con):
    expr1 = L("Hello, World!").re_replace(".", r"\0\0")
    expr2 = L("Hello, World!").re_replace("^", "here: ")

    assert con.execute(expr1) == "HHeelllloo,,  WWoorrlldd!!"
    assert con.execute(expr2) == "here: Hello, World!"


@pytest.mark.parametrize(
    ("value", "expected"),
    [(L("a"), 0), (L("b"), 1), (L("d"), -1)],  # TODO: what's the expected?
)
def test_find_in_set(con, value, expected):
    vals = list("abc")
    expr = value.find_in_set(vals)
    assert con.execute(expr) == expected


def test_string_column_find_in_set(con, alltypes, snapshot):
    s = alltypes.string_col
    vals = list("abc")

    expr = s.find_in_set(vals)
    snapshot.assert_match(expr.compile(), "out.sql")
    assert len(con.execute(expr))


@pytest.mark.parametrize(
    ("expr", "expected"),
    [
        (L("foobar").find("bar"), 3),
        (L("foobar").find("baz"), -1),
        (L("foobar").like("%bar"), True),
        (L("foobar").like("foo%"), True),
        (L("foobar").like("%baz%"), False),
        (L("foobar").like(["%bar"]), True),
        (L("foobar").like(["foo%"]), True),
        (L("foobar").like(["%baz%"]), False),
        (L("foobar").like(["%bar", "foo%"]), True),
        (L("foobarfoo").replace("foo", "H"), "HbarH"),
    ],
)
def test_string_find_like(con, expr, expected):
    assert con.execute(expr) == expected


def test_string_column_like(con, alltypes, snapshot):
    expr = alltypes.string_col.like("foo%")
    snapshot.assert_match(expr.compile(), "out1.sql")
    assert len(con.execute(expr))

    expr = alltypes.string_col.like(["foo%", "%bar"])
    snapshot.assert_match(expr.compile(), "out2.sql")
    assert len(con.execute(expr))


def test_string_column_find(con, alltypes, snapshot):
    s = alltypes.string_col

    expr = s.find("a")
    snapshot.assert_match(expr.compile(), "out1.sql")
    assert len(con.execute(expr))

    expr = s.find(s)
    snapshot.assert_match(expr.compile(), "out2.sql")
    assert len(con.execute(expr))


@pytest.mark.parametrize(
    "call",
    [
        param(methodcaller("log"), id="log"),
        param(methodcaller("log2"), id="log2"),
        param(methodcaller("log10"), id="log10"),
        param(methodcaller("round"), id="round"),
        param(methodcaller("round", 0), id="round_0"),
        param(methodcaller("round", 2), id="round_2"),
        param(methodcaller("exp"), id="exp"),
        param(methodcaller("abs"), id="abs"),
        param(methodcaller("ceil"), id="ceil"),
        param(methodcaller("sqrt"), id="sqrt"),
        param(methodcaller("sign"), id="sign"),
    ],
)
def test_translate_math_functions(con, alltypes, call, snapshot):
    expr = call(alltypes.double_col)
    snapshot.assert_match(expr.compile(), "out.sql")
    assert len(con.execute(expr))


@pytest.mark.parametrize(
    ("expr", "expected"),
    [
        pytest.param(L(-5).abs(), 5, id="abs_neg"),
        pytest.param(L(5).abs(), 5, id="abs"),
        pytest.param(L(5.5).round(), 6.0, id="round"),
        pytest.param(L(5.556).round(2), 5.56, id="round_places"),
        pytest.param(L(5.556).ceil(), 6.0, id="ceil"),
        pytest.param(L(5.556).floor(), 5.0, id="floor"),
        pytest.param(L(5.556).sign(), 1, id="sign"),
        pytest.param(L(-5.556).sign(), -1, id="sign_neg"),
        pytest.param(L(0).sign(), 0, id="sign_zero"),
        pytest.param(L(5.556).sqrt(), math.sqrt(5.556), id="sqrt"),
        pytest.param(L(5.556).log(2), math.log(5.556, 2), id="log2_arg"),
        pytest.param(L(5.556).log2(), math.log(5.556, 2), id="log2"),
        pytest.param(L(5.556).log10(), math.log10(5.556), id="log10"),
        # clickhouse has different functions for exp/ln that are faster
        # than the defaults, but less precise
        #
        # we can't use the e() function as it still gives different results
        # from `math.exp`
        pytest.param(
            L(5.556).exp().round(8),
            round(math.exp(5.556), 8),
            id="exp",
        ),
        pytest.param(
            L(5.556).ln().round(7),
            round(math.log(5.556), 7),
            id="ln",
        ),
    ],
)
def test_math_functions(con, expr, expected):
    assert con.execute(expr) == expected


def test_greatest_least(con, alltypes, snapshot):
    expr = ibis.greatest(alltypes.int_col, 10)
    snapshot.assert_match(expr.compile(), "out1.sql")
    assert len(con.execute(expr))

    expr = ibis.greatest(alltypes.int_col, alltypes.bigint_col)
    snapshot.assert_match(expr.compile(), "out2.sql")
    assert len(con.execute(expr))

    expr = ibis.least(alltypes.int_col, 10)
    snapshot.assert_match(expr.compile(), "out3.sql")
    assert len(con.execute(expr))

    expr = ibis.least(alltypes.int_col, alltypes.bigint_col)
    snapshot.assert_match(expr.compile(), "out4.sql")
    assert len(con.execute(expr))


@pytest.mark.parametrize(
    ("expr", "expected"),
    [
        (L("abcd").re_search("[a-z]"), True),
        (L("abcd").re_search(r"[\d]+"), False),
        (L("1222").re_search(r"[\d]+"), True),
    ],
)
def test_regexp(con, expr, expected):
    assert con.execute(expr) == expected


@pytest.mark.parametrize(
    ("expr", "expected"),
    [
        param(L("abcd").re_extract("([a-z]+)", 0), "abcd", id="simple"),
        # (L('abcd').re_extract('(ab)(cd)', 1), 'cd'),
        # valid group number but no match => None
        param(L("abcd").re_extract(r"(\\d)", 0), None, id="valid_group_no_match"),
        # match but not a valid group number => NULL
        param(L("abcd").re_extract("abcd", 3), None, id="invalid_group_match"),
    ],
)
def test_regexp_extract(con, expr, expected):
    assert con.execute(expr) == expected


def test_column_regexp_extract(con, alltypes, snapshot):
    expr = alltypes.string_col.re_extract(r"[\d]+", 3)
    snapshot.assert_match(expr.compile(), "out.sql")
    assert len(con.execute(expr))


def test_column_regexp_replace(con, alltypes, snapshot):
    expr = alltypes.string_col.re_replace(r"[\d]+", "aaa")
    snapshot.assert_match(expr.compile(), "out.sql")
    assert len(con.execute(expr))


def test_numeric_builtins_work(alltypes, df):
    expr = alltypes.double_col
    result = expr.execute()
    expected = df.double_col.fillna(0)
    tm.assert_series_equal(result, expected)


def test_null_column(alltypes):
    t = alltypes
    nrows = t.count().execute()
    expr = t.mutate(na_column=ibis.NA).na_column
    result = expr.execute()
    expected = pd.Series([None] * nrows, name="na_column")
    tm.assert_series_equal(result, expected)


def test_literal_none_to_nullable_column(alltypes):
    # GH: 2985
    t = alltypes
    nrows = t.count().execute()
    expr = t.mutate(
        ibis.literal(None, dt.String(nullable=True)).name("nullable_string_column")
    )
    result = expr["nullable_string_column"].execute()
    expected = pd.Series([None] * nrows, name="nullable_string_column")
    tm.assert_series_equal(result, expected)


def test_timestamp_from_integer(con, alltypes, snapshot):
    # timestamp_col has datetime type
    expr = alltypes.int_col.to_timestamp()
    snapshot.assert_match(expr.compile(), "out.sql")
    assert len(con.execute(expr))


def test_count_distinct_with_filter(alltypes):
    expr = alltypes.string_col.nunique(where=alltypes.string_col.cast("int64") > 1)
    result = expr.execute()
    expected = alltypes.string_col.execute()
    expected = expected[expected.astype("int64") > 1].nunique()
    assert result == expected


@pytest.mark.parametrize(
    ("sep", "where_case"),
    [
        param(",", None, id="comma_none"),
        param("-", None, id="minus_none"),
        param(",", 0, id="comma_zero"),
    ],
)
def test_group_concat(alltypes, sep, where_case, snapshot):
    where = None if where_case is None else alltypes.bool_col == where_case
    expr = alltypes.string_col.group_concat(sep, where)
    snapshot.assert_match(expr.compile(), "out.sql")


def test_hash(alltypes, snapshot):
    expr = alltypes.string_col.hash()
    snapshot.assert_match(expr.compile(), "out.sql")


def test_udf_in_array_map(alltypes):
    @udf.scalar.builtin(name="plus")
    def my_add(a: int, b: int) -> int:
        ...

    n = 5
    expr = (
        alltypes[alltypes.int_col == 1]
        .limit(n)
        .int_col.collect()
        .map(lambda x: my_add(x, 1))
    )
    result = expr.execute()
    assert result == [2] * n


def test_udf_in_array_filter(alltypes):
    @udf.scalar.builtin(name="equals")
    def my_eq(a: int, b: int) -> bool:
        ...

    expr = alltypes.int_col.collect().filter(lambda x: my_eq(x, 1))
    result = expr.execute()
    assert set(result) == {1}
